import os
import tensorflow as tf
from tensorflow.python.util import nest
import numpy as np

curr_dir = os.path.dirname(os.path.realpath(__file__))

from qiznlp.common.modules.common_layers import shape_list, mask_nonpad_from_embedding
from qiznlp.common.modules.embedding import embedding
from qiznlp.common.modules.encoder import conv_multi_kernel, batch_coattention_nnsubmulti
from qiznlp.common.modules.encoder import multihead_attention_encoder as multihead_attention
import qiznlp.common.utils as utils

conf = utils.dict2obj({
    'max_turn': 4,
    'max_char': 4,
    'vocab_size': 13124,
    'embed_size': 300,
    'char_vocab_size': 2943,
    'char_embed_size': 300,
    'hidden_size': 300,
    'num_heads': 6,
    'dropout_rate': 0.2,
    'lr': 1e-3,
    'pretrain_emb': None,
})


class Model(object):
    def __init__(self, build_graph=True, **kwargs):
        self.conf = conf
        self.run_model = kwargs.get('run_model', None)  # acquire outside run_model instance
        if build_graph:
            # build placeholder
            self.build_placeholder()
            # build model
            self.model_name = kwargs.get('model_name', 'MRFN_1')
            {
                # Multi-Turn Response Selection for Chatbots with Deep Attention Matching Network https://github.com/julianser/hed-dlg
                'DAM': self.build_model1,  # https://www.aclweb.org/anthology/P18-1103/
                # Multi-Representation Fusion Network for Multi-Turn Response Selection in Retrieval-Based Chatbots https://github.com/chongyangtao/MRFN
                'MRFN': self.build_model2,  # https://dl.acm.org/doi/10.1145/3289600.3290985
                # personal modify: use word_char fuse as base embedding, but seem not convergent
                'MRFN_1': self.build_model3,
                # add new here
            }[self.model_name]()
            print(f'model_name: {self.model_name} build graph ok!')

    def build_placeholder(self):
        # placeholder
        # 原则上模型输入输出不变，不需换新model
        self.multi_s1 = tf.placeholder(tf.int32, [None, None, None], name='multi_s1')  # [batch,turn,len]
        self.s2 = tf.placeholder(tf.int32, [None, None], name='s2')  # [batch,len]

        # mrfn use char
        self.char_multi_s1 = tf.placeholder(tf.int32, [None, None, None, None], name='char_multi_s1')  # [batch,turn,len,char]
        self.char_s2 = tf.placeholder(tf.int32, [None, None, None], name='char_s2')  # [batch,len,char]

        self.target = tf.placeholder(tf.int32, [None], name='target')
        self.dropout_rate = tf.placeholder(tf.float32, name="dropout_rate")

    def build_model1(self):
        from qiznlp.common.modules.DAM import layers as layers
        from qiznlp.common.modules.DAM import operations as op

        fix_turn = conf.max_turn  # 4  # dam需显式指定s1的turn 因为使用了tf.unstack
        fix_len = 50  # dam需显式指定s1和s2的len 因为与最后立方体卷积的参数相关

        # 方式1:通过重写原来的placeholder, 此时要求输入数据符合shape, 不够灵活
        # self.multi_s1 = tf.placeholder(tf.int32, [None, 4, 50], name='multi_s1')  # [batch,turn,len]
        # self.s2 = tf.placeholder(tf.int32, [None, 50], name='s2')  # [batch,len]
        # multi_s1 = self.multi_s1
        # s2 = self.s2

        # 方式2:通过在原来的placeholder补齐来实现
        batch_size, s1_cur_turn, s1_cur_len = shape_list(self.multi_s1)
        multi_s1_pad_len = tf.zeros([batch_size, s1_cur_turn, fix_len - s1_cur_len], dtype=tf.int32)  # [batch,turn,-]
        multi_s1 = tf.concat([self.multi_s1, multi_s1_pad_len], -1)  # [batch,turn,50]
        multi_s1_pad_turn = tf.zeros([batch_size, fix_turn - s1_cur_turn, fix_len], dtype=tf.int32)  # [batch,-,50]
        multi_s1 = tf.concat([multi_s1, multi_s1_pad_turn], 1)  # [batch,4,50]
        multi_s1.set_shape([None, fix_turn, fix_len])

        s2_cur_len = shape_list(self.s2)[1]
        s2_pad = tf.zeros([batch_size, fix_len - s2_cur_len], dtype=tf.int32)  # [batch,-]
        s2 = tf.concat([self.s2, s2_pad], -1)  # [batch,50]
        s2.set_shape([None, fix_len])

        embed_size = conf.embed_size
        vocab_size = conf.vocab_size
        stack_num = 5

        is_positional = True

        batch_size, s1_turn, s1_length = shape_list(multi_s1)
        # ====== multi_s1 embedding
        # multi_s1_embed: [batch,turn,len,emb]
        multi_s1_embed, _ = embedding(tf.expand_dims(multi_s1, -1), vocab_size, embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        multi_s1_mask = mask_nonpad_from_embedding(multi_s1_embed)  # [batch,turn,len] 1 for nonpad; 0 for pad
        multi_s1_seqlen = tf.cast(tf.reduce_sum(multi_s1_mask, -1), tf.int32)  # [batch,turn]

        # ====== s2 embedding
        # s2_embed: [batch,len,emb] 
        s2_embed, _ = embedding(tf.expand_dims(s2, -1), vocab_size, embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, -1), tf.int32)  # [batch]

        # s2
        Hr = s2_embed
        if is_positional and stack_num > 0:
            with tf.variable_scope('positional'):
                Hr = op.positional_encoding_vector(Hr, max_timescale=10)
        Hr = tf.layers.dropout(Hr, rate=self.dropout_rate)
        Hr_stack = [Hr]

        for index in range(stack_num):
            with tf.variable_scope('self_stack_%d' % index):
                Hr = layers.block(
                    Hr, Hr, Hr,
                    Q_lengths=s2_seqlen,
                    K_lengths=s2_seqlen,
                    drop_prob=self.dropout_rate)  # attention dropout
                Hr_stack.append(Hr)

        # multi_s1
        s1_embed_lst = tf.unstack(multi_s1_embed, num=s1_turn, axis=1)  # turn * [batch,len,emb]
        s1_seqlen_lst = tf.unstack(multi_s1_seqlen, num=s1_turn, axis=1)  # turn * [batch]

        sim_turns = []
        # for every turn calculate matching vector
        for s1_embed, s1_seqlen in zip(s1_embed_lst, s1_seqlen_lst):
            Hu = s1_embed  # [batch,len,emb]

            if is_positional and stack_num > 0:
                with tf.variable_scope('positional', reuse=True):
                    Hu = op.positional_encoding_vector(Hu, max_timescale=10)
            Hu = tf.layers.dropout(Hu, rate=self.dropout_rate)
            Hu_stack = [Hu]

            for index in range(stack_num):
                with tf.variable_scope('self_stack_%d' % index, reuse=True):
                    Hu = layers.block(
                        Hu, Hu, Hu,
                        Q_lengths=s1_seqlen,
                        K_lengths=s1_seqlen,
                        drop_prob=self.dropout_rate)  # attention dropout

                    Hu_stack.append(Hu)

            r_a_t_stack = []
            t_a_r_stack = []
            for index in range(stack_num + 1):
                with tf.variable_scope('t_attend_r_%d' % index, reuse=tf.AUTO_REUSE):
                    t_a_r = layers.block(
                        Hu_stack[index], Hr_stack[index], Hr_stack[index],
                        Q_lengths=s1_seqlen,
                        K_lengths=s2_seqlen,
                        drop_prob=self.dropout_rate)  # attention dropout

                with tf.variable_scope('r_attend_t_%d' % index, reuse=tf.AUTO_REUSE):
                    r_a_t = layers.block(
                        Hr_stack[index], Hu_stack[index], Hu_stack[index],
                        Q_lengths=s2_seqlen,
                        K_lengths=s1_seqlen,
                        drop_prob=self.dropout_rate)  # attention dropout

                t_a_r_stack.append(t_a_r)
                r_a_t_stack.append(r_a_t)

            t_a_r_stack.extend(Hu_stack)
            r_a_t_stack.extend(Hr_stack)

            t_a_r = tf.stack(t_a_r_stack, axis=-1)
            r_a_t = tf.stack(r_a_t_stack, axis=-1)

            # calculate similarity matrix
            with tf.variable_scope('similarity'):
                # sim shape [batch, turn, len, 2*stack_num+1]
                # divide sqrt(200) to prevent gradient explosion
                sim = tf.einsum('biks,bjks->bijs', t_a_r, r_a_t) / tf.sqrt(200.0)

            sim_turns.append(sim)

        # cnn and aggregation
        sim = tf.stack(sim_turns, axis=1)
        print('sim shape: %s' % sim.shape)
        with tf.variable_scope('cnn_aggregation'):
            final_info = layers.CNN_3d(sim, 32, 16)
            # for douban
            # final_info = layers.CNN_3d(sim, 16, 16)

        self.loss, self.logits = layers.loss(final_info, self.target)

        y_prob = tf.nn.sigmoid(self.logits)
        self.y_prob = tf.identity(y_prob, 'y_prob')

        with tf.name_scope("accuracy"):
            self.correct = tf.equal(
                tf.cast(tf.greater_equal(self.y_prob, 0.5), tf.int32),
                self.target)
            self.accuracy = tf.reduce_mean(tf.cast(self.correct, tf.float32))

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)

        grads_and_vars = self.optimizer.compute_gradients(self.loss)
        capped_gvs = [(tf.clip_by_value(grad, -1, 1), var) for grad, var in grads_and_vars]
        self.train_op = self.optimizer.apply_gradients(capped_gvs, global_step=self.global_step)

        # self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model2(self):

        def interaction_matching_batch(s1_hidden, s2_hidden, hidden_size, s1_mask, s1_turn, scope="interaction_matching", reuse=None):
            # s1_hidden [batch*turn,len,emb]
            # s2_hidden [batch*turn,len,emb]
            with tf.variable_scope(scope, reuse=reuse):
                res_coatt = batch_coattention_nnsubmulti(s1_hidden,
                                                         s2_hidden,
                                                         tf.to_float(s1_mask),
                                                         scope='%s_res' % scope)

                res_coatt = tf.layers.dropout(res_coatt, self.dropout_rate)  # [batch*turn,len,hid]

                # 第一级rnn 规约句子
                res_coatt_cell = tf.contrib.rnn.GRUCell(hidden_size)
                _, res_final = tf.nn.dynamic_rnn(res_coatt_cell, res_coatt, dtype=tf.float32, scope='first_level_gru')

                res_feature = res_final  # [batch*turn,hid]
                res_feature = tf.reshape(res_feature, [-1, s1_turn, hidden_size])  # [batch,turn,hid]

                # 第二级rnn 规约多轮
                final_gru_cell = tf.contrib.rnn.GRUCell(hidden_size)
                _, last_hidden = tf.nn.dynamic_rnn(final_gru_cell, res_feature, dtype=tf.float32, scope='final_gru')
                logits = tf.layers.dense(last_hidden, 2, kernel_initializer=tf.contrib.layers.xavier_initializer(), name='final_v')
                return logits

        self.loss_lst = []
        self.y_prob_lst = []

        use_char = True
        use_word = True
        use_seq = True
        use_conv = True
        use_self = True
        use_cross = True

        embed_size = conf.embed_size
        vocab_size = conf.vocab_size
        cembed_size = conf.char_embed_size
        cvocab_size = conf.char_vocab_size

        batch_size, s1_turn, s1_length = shape_list(self.multi_s1)
        char_len = shape_list(self.char_multi_s1)[3]  # 4 here

        # ====== multi_s1 embedding
        # _norm mean reshape to batch * turn
        # multi_s1_embed: [batch,turn,len,emb]
        multi_s1_embed, _ = embedding(tf.expand_dims(self.multi_s1, -1), vocab_size, embed_size, name='share_word_embedding', pretrain_embedding=conf.pretrain_emb)
        multi_s1_mask = mask_nonpad_from_embedding(multi_s1_embed)  # [batch,turn,len] 1 for nonpad; 0 for pad
        multi_s1_mask_norm = tf.reshape(multi_s1_mask, [-1, s1_length])  # [batch*turn,len]
        multi_s1_seqlen = tf.cast(tf.reduce_sum(multi_s1_mask, -1), tf.int32)  # [batch,turn]
        multi_s1_seqlen_norm = tf.reshape(multi_s1_seqlen, [-1])  # [batch*turn]
        multi_s1_embed_norm = tf.reshape(multi_s1_embed, [-1, s1_length, embed_size])  # [batch*turn,len,emb]

        # ====== s2 embedding
        s2_length = shape_list(self.s2)[1]
        # _turn mean s2 expand and tile to batch,turn,..
        # _norm mean s2 reshape to batch*turn
        # s2_embed: [batch,len,emb] 
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), vocab_size, embed_size, name='share_word_embedding', pretrain_embedding=conf.pretrain_emb)
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_mask_turn = tf.tile(tf.expand_dims(s2_mask, 1), [1, s1_turn, 1])  # [batch,turn,len]
        s2_mask_norm = tf.reshape(s2_mask_turn, [-1, s2_length])  # [batch*turn,len]
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, -1), tf.int32)  # [batch]
        s2_embed_turn = tf.tile(tf.expand_dims(s2_embed, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,emb]
        s2_embed_norm = tf.reshape(s2_embed_turn, [-1, s2_length, embed_size])  # [batch*turn,len,emb]

        # ====== representation and interaction

        # ====== char representation
        if use_char:
            char_filter = 100
            kernels_size = [2, 3, 4]

            # ====== multi_s1 char embedding
            # _norm mean reshape to batch*turn
            # s1_char_embed: [batch,turn,len,char,emb]
            char_multi_s1_embed, _ = embedding(tf.expand_dims(self.char_multi_s1, -1), cvocab_size, cembed_size, name='share_char_embedding')
            char_multi_s1_mask = mask_nonpad_from_embedding(char_multi_s1_embed)  # [batch,turn,len,4] 1 for nonpad; 0 for pad
            char_multi_s1_embed = tf.reshape(char_multi_s1_embed, [-1, char_len, cembed_size])  # [batch*turn*len,4,emb]
            char_multi_s1_conv = conv_multi_kernel(char_multi_s1_embed, char_filter, kernels_size=kernels_size,
                                                   padding='same', use_bias=True, activation=tf.nn.relu,
                                                   name="char_conv", layer_norm_finally=False)  # [batch*turn*len,4,filter*3]
            # remove_pad 将pad的vector变为极小
            char_multi_s1_conv += tf.expand_dims(tf.reshape(1. - char_multi_s1_mask, [-1, char_len]), -1) * -1e9  # [batch*turn*len,4,1]
            # max-pooling
            char_multi_s1_conv = tf.reduce_max(char_multi_s1_conv, axis=1)  # [batch*turn*len,hid]
            char_multi_s1_conv = tf.reshape(char_multi_s1_conv, [-1, s1_turn, s1_length, char_filter * len(kernels_size)])  # [batch,turn,len,hid]
            char_multi_s1_embed = tf.layers.dropout(char_multi_s1_conv, self.dropout_rate)
            char_multi_s1_embed_norm = tf.reshape(char_multi_s1_embed, [-1, s1_length, char_filter * len(kernels_size)])  # [batch*turn,len,hid]

            # ====== s2 char embedding
            # _turn mean expand and tile to batch,turn,..
            # _norm mean reshape to batch*turn
            # char_s2_embed: [batch,len,4,emb]
            char_s2_embed, _ = embedding(tf.expand_dims(self.char_s2, -1), cvocab_size, cembed_size, name='share_char_embedding')
            char_s2_mask = mask_nonpad_from_embedding(char_s2_embed)  # [batch,len,4] 1 for nonpad; 0 for pad
            char_s2_embed = tf.reshape(char_s2_embed, [-1, char_len, cembed_size])  # [batch*len,4,emb]
            char_s2_conv = conv_multi_kernel(char_s2_embed, char_filter, kernels_size=kernels_size,
                                             padding='same', use_bias=True, activation=tf.nn.relu,
                                             name="char_conv", layer_norm_finally=False)  # [batch*len,4,filter*3]
            # remove_pad 将pad的vector变为极小
            char_s2_conv += tf.expand_dims(tf.reshape(1. - char_s2_mask, [-1, char_len]), -1) * -1e9  # [batch*len,4,1]
            # max-pooling
            char_s2_conv = tf.reduce_max(char_s2_conv, axis=1)  # [batch*len,hid]
            char_s2_conv = tf.reshape(char_s2_conv, [-1, s2_length, char_filter * len(kernels_size)])  # [batch,len,hid]
            char_s2_embed = tf.layers.dropout(char_s2_conv, self.dropout_rate)
            char_s2_embed_turn = tf.tile(tf.expand_dims(char_s2_embed, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,hid]
            char_s2_embed_norm = tf.reshape(char_s2_embed_turn, [-1, s2_length, char_filter * len(kernels_size)])  # [batch*turn,len,hid]

            char_interaction = interaction_matching_batch(char_multi_s1_embed_norm,
                                                          char_s2_embed_norm,
                                                          embed_size,
                                                          multi_s1_mask_norm,
                                                          s1_turn,
                                                          scope='char_interaction_matching')
            loss_char = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=char_interaction))
            self.loss_lst.append(loss_char)
            self.y_prob_lst.append(tf.nn.softmax(char_interaction))

        # ====== word representation
        if use_word:
            word_interaction = interaction_matching_batch(multi_s1_embed_norm,
                                                          s2_embed_norm,
                                                          embed_size,
                                                          multi_s1_mask_norm,
                                                          s1_turn,
                                                          scope='word_interaction_matching')
            loss_word = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=word_interaction))
            self.loss_lst.append(loss_word)
            self.y_prob_lst.append(tf.nn.softmax(word_interaction))

        # ====== rnn representation
        rnn_dim = 300
        if use_seq:
            with tf.variable_scope("rnn_embeddings"):
                sentence_gru_cell = tf.contrib.rnn.GRUCell(rnn_dim)
                with tf.variable_scope('sentence_gru'):
                    # [batch*turn,len,hid]
                    s1_gru_hidden_norm, _ = tf.nn.dynamic_rnn(sentence_gru_cell, multi_s1_embed_norm, sequence_length=multi_s1_seqlen_norm, dtype=tf.float32)

                with tf.variable_scope('sentence_gru', reuse=True):
                    # [batch,len,hid]
                    s2_gru_hidden, _ = tf.nn.dynamic_rnn(sentence_gru_cell, s2_embed, sequence_length=s2_seqlen, dtype=tf.float32)

            s2_gru_hidden_turn = tf.tile(tf.expand_dims(s2_gru_hidden, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,hid]
            s2_gru_hidden_norm = tf.reshape(s2_gru_hidden_turn, [-1, s2_length, rnn_dim])  # [batch*turn,len,hid]

            seq_interaction = interaction_matching_batch(s1_gru_hidden_norm,
                                                         s2_gru_hidden_norm,
                                                         rnn_dim,
                                                         multi_s1_mask_norm,
                                                         s1_turn,
                                                         scope='seg_interaction_matching')
            loss_seq = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=seq_interaction))
            self.loss_lst.append(loss_seq)
            self.y_prob_lst.append(tf.nn.softmax(seq_interaction))

        # ====== conv representation
        if use_conv:
            conv_dim = 100
            kernels = [1, 2, 3, 4]
            with tf.variable_scope("conv_embeddings"):
                s1_conv_hidden_norm = conv_multi_kernel(multi_s1_embed_norm, conv_dim, kernels_size=kernels,
                                                        padding='SAME', use_bias=True, activation=tf.nn.relu,
                                                        name="seq_conv", layer_norm_finally=True)  # [batch*turn,len,filter*4]

                s2_conv_hidden = conv_multi_kernel(s2_embed, conv_dim, kernels_size=kernels,
                                                   padding='SAME', use_bias=True, activation=tf.nn.relu,
                                                   name="seq_conv", layer_norm_finally=True)  # [batch,len,filter*4]

                s2_conv_hidden_turn = tf.tile(tf.expand_dims(s2_conv_hidden, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,filter*4]
                s2_conv_hidden_norm = tf.reshape(s2_conv_hidden_turn, [-1, s2_length, conv_dim * len(kernels)])  # [batch*turn,len,filter*4]

            conv_interaction = interaction_matching_batch(s1_conv_hidden_norm,
                                                          s2_conv_hidden_norm,
                                                          conv_dim * len(kernels),
                                                          multi_s1_mask_norm,
                                                          s1_turn,
                                                          scope='conv_interaction_matching')
            loss_conv = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=conv_interaction))
            self.loss_lst.append(loss_conv)
            self.y_prob_lst.append(tf.nn.softmax(conv_interaction))

        # ====== self attention
        if use_self:
            with tf.variable_scope("self_att_embeddings"):
                s1_self_att_hidden_norm = multihead_attention(multi_s1_embed_norm,
                                                              multi_s1_embed_norm,
                                                              multi_s1_mask_norm,
                                                              embed_size,
                                                              conf.num_heads,
                                                              self.dropout_rate)

                s2_self_att_hidden = multihead_attention(s2_embed,
                                                         s2_embed,
                                                         s2_mask,
                                                         embed_size,
                                                         conf.num_heads,
                                                         self.dropout_rate)

                s2_self_att_hidden_turn = tf.tile(tf.expand_dims(s2_self_att_hidden, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,hid]
                s2_self_att_hidden_norm = tf.reshape(s2_self_att_hidden_turn, [-1, s2_length, embed_size])  # [batch*turn,len,hid]

            self_att_interaction = interaction_matching_batch(s1_self_att_hidden_norm,
                                                              s2_self_att_hidden_norm,
                                                              embed_size,
                                                              multi_s1_mask_norm,
                                                              s1_turn,
                                                              scope='self_att_interaction_matching')
            loss_self = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=self_att_interaction))
            self.loss_lst.append(loss_self)
            self.y_prob_lst.append(tf.nn.softmax(self_att_interaction))

        # cross attention
        if use_cross:
            with tf.variable_scope("cross_att_embeddings"):
                s1_cross_att_hidden_norm = multihead_attention(multi_s1_embed_norm,
                                                               s2_embed_norm,
                                                               s2_mask_norm,
                                                               embed_size,
                                                               conf.num_heads,
                                                               self.dropout_rate)

                s2_cross_attn_hidden_norm = multihead_attention(s2_embed_norm,
                                                                multi_s1_embed_norm,
                                                                multi_s1_mask_norm,
                                                                embed_size,
                                                                conf.num_heads,
                                                                self.dropout_rate)

                # s2_cross_attn_hidden_lst = []
                # self.response_cross_att_embeddings = []
                # s1_embed_lst = tf.unstack(multi_s1_embed, axis=1)  # turn * [batch,len,emb]
                # s1_mask_lst = tf.unstack(multi_s1_mask, axis=1)  # turn * [batch,len]
                # for k, s1_embed_ in enumerate(s1_embed_lst):
                #     s2_cross_att_hidden = multihead_attention(s2_embed,
                #                                               s1_embed_,
                #                                               s1_mask_lst[k],
                #                                               embed_size,
                #                                               conf.num_heads,
                #                                               self.dropout_rate)
                #     s2_cross_attn_hidden_lst.append(s2_cross_att_hidden)  # turn * [batch,len,hid]
                # 
                # s2_cross_attn_hidden_turn = tf.stack(s2_cross_attn_hidden_lst, axis=1)  # [batch,turn,len,hid]
                # s2_cross_attn_hidden_norm = tf.reshape(s2_cross_attn_hidden_turn, [-1, s2_length, embed_size])

            cross_att_interaction = interaction_matching_batch(s1_cross_att_hidden_norm,
                                                               s2_cross_attn_hidden_norm,
                                                               embed_size,
                                                               multi_s1_mask_norm,
                                                               s1_turn,
                                                               scope='cross_att_interaction_matching')
            loss_cross = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=cross_att_interaction))
            self.loss_lst.append(loss_cross)
            self.y_prob_lst.append(tf.nn.softmax(cross_att_interaction))

        self.loss = sum(self.loss_lst)
        y_prob = sum(self.y_prob_lst)  # [batch,2]
        y_prob = y_prob[:, 1]
        self.y_prob = tf.identity(y_prob, 'y_prob')

        with tf.name_scope("accuracy"):
            self.correct = tf.equal(
                tf.cast(tf.greater_equal(self.y_prob, 3), tf.int32),  # 因为加了6个0-1的概率，随意中值用2.5
                self.target)
            self.accuracy = tf.reduce_mean(tf.cast(self.correct, tf.float32))

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model3(self):

        def interaction_matching_batch(s1_hidden, s2_hidden, hidden_size, s1_mask, s1_turn, scope="interaction_matching", reuse=None):
            # s1_hidden [batch*turn,len,emb]
            # s2_hidden [batch*turn,len,emb]
            with tf.variable_scope(scope, reuse=reuse):
                res_coatt = batch_coattention_nnsubmulti(s1_hidden,
                                                         s2_hidden,
                                                         tf.to_float(s1_mask),
                                                         scope='%s_res' % scope)

                res_coatt = tf.layers.dropout(res_coatt, self.dropout_rate)  # [batch*turn,len,hid]

                # 第一级rnn 规约句子
                res_coatt_cell = tf.contrib.rnn.GRUCell(hidden_size)
                _, res_final = tf.nn.dynamic_rnn(res_coatt_cell, res_coatt, dtype=tf.float32, scope='first_level_gru')

                res_feature = res_final  # [batch*turn,hid]
                res_feature = tf.reshape(res_feature, [-1, s1_turn, hidden_size])  # [batch,turn,hid]

                # 第二级rnn 规约多轮
                final_gru_cell = tf.contrib.rnn.GRUCell(hidden_size)
                _, last_hidden = tf.nn.dynamic_rnn(final_gru_cell, res_feature, dtype=tf.float32, scope='final_gru')
                logits = tf.layers.dense(last_hidden, 2, kernel_initializer=tf.contrib.layers.xavier_initializer(), name='final_v')
                return logits

        self.loss_lst = []
        self.y_prob_lst = []

        use_word = True
        use_seq = True
        use_conv = True
        use_self = True
        use_cross = True

        embed_size = conf.embed_size
        vocab_size = conf.vocab_size
        cembed_size = conf.char_embed_size
        cvocab_size = conf.char_vocab_size

        batch_size, s1_turn, s1_length = shape_list(self.multi_s1)
        char_len = shape_list(self.char_multi_s1)[3]  # 4 here

        # ====== multi_s1 embedding
        # norm mean reshape to batch * turn
        # multi_s1_embed: [batch,turn,len,emb]
        multi_s1_embed, _ = embedding(tf.expand_dims(self.multi_s1, -1), vocab_size, embed_size, name='share_word_embedding', pretrain_embedding=conf.pretrain_emb)
        multi_s1_mask = mask_nonpad_from_embedding(multi_s1_embed)  # [batch,turn,len] 1 for nonpad; 0 for pad
        multi_s1_mask_norm = tf.reshape(multi_s1_mask, [-1, s1_length])  # [batch*turn,len]
        multi_s1_seqlen = tf.cast(tf.reduce_sum(multi_s1_mask, -1), tf.int32)  # [batch,turn]
        multi_s1_seqlen_norm = tf.reshape(multi_s1_seqlen, [-1])  # [batch*turn]

        # ====== multi_s1 char embedding
        # s1_char_embed: [batch,turn,len,char,emb]
        char_multi_s1_embed, _ = embedding(tf.expand_dims(self.char_multi_s1, -1), cvocab_size, cembed_size, name='share_char_embedding')
        char_multi_s1_mask = mask_nonpad_from_embedding(char_multi_s1_embed)  # [batch,turn,len,4] 1 for nonpad; 0 for pad
        char_multi_s1_embed = tf.reshape(char_multi_s1_embed, [-1, char_len, cembed_size])  # [batch*turn*len,4,emb]
        
        # 方式1:char cnn + max-pooling
        # char_filter = 100
        # kernels_size = [2, 3, 4]
        # char_multi_s1_conv = conv_multi_kernel(char_multi_s1_embed, char_filter, kernels_size=kernels_size,
        #                                        padding='same', use_bias=True, activation=tf.nn.relu,
        #                                        name="char_conv", layer_norm_finally=False)  # [batch*turn*len,4,filter*3]
        # # max-pooling
        # # remove_pad 将pad的vector变为极小
        # char_multi_s1_conv += tf.expand_dims(tf.reshape(1. - char_multi_s1_mask, [-1, char_len]), -1) * -1e9  # [batch*turn*len,4,1]
        # char_multi_s1_conv = tf.reduce_max(char_multi_s1_conv, axis=1)  # [batch*turn*len,hid]
        # char_multi_s1_conv = tf.reshape(char_multi_s1_conv, [-1, s1_turn, s1_length, char_filter * len(kernels_size)])  # [batch,turn,len,hid]
        # char_multi_s1_conv = tf.layers.dropout(char_multi_s1_conv, self.dropout_rate)
        
        # 方式2:char mean-pooling
        char_multi_s1_sum = tf.reduce_sum(char_multi_s1_embed, axis=-2)  # [batch*turn*len,hid]
        char_multi_s1_num = tf.reduce_sum(tf.reshape(char_multi_s1_mask, [-1, char_len]), -1, keepdims=True)  # [batch*turn*len,1]
        char_multi_s1_mean = char_multi_s1_sum / char_multi_s1_num  #   # [batch*turn*len,hid]
        
        char_multi_s1_mean = tf.reshape(char_multi_s1_mean, [-1, s1_turn, s1_length, cembed_size])  # [batch,turn,len,hid]
        # char_multi_s1_mean = tf.layers.dropout(char_multi_s1_mean, self.dropout_rate)

        # ====== fuse word and char embedding for multi_s1
        multi_s1_embed += char_multi_s1_mean  # seem not convergence -_-
        # multi_s1_embed = tf.concat([multi_s1_embed, char_multi_s1_mean], axis=-1)  # [batch,turn,len,emb+]
        # [batch,turn,len,300]
        # multi_s1_embed = tf.layers.dense(multi_s1_embed, 300, use_bias=False, kernel_initializer=tf.contrib.layers.xavier_initializer(), name='word_fuse', reuse=tf.AUTO_REUSE)
        multi_s1_embed_norm = tf.reshape(multi_s1_embed, [-1, s1_length, 300])  # [batch*turn,len,300]

        # ====== s2 embedding
        s2_length = shape_list(self.s2)[1]
        # _turn mean s2 expand and tile to batch,turn,..
        # _norm mean s2 reshape to batch*turn
        # s2_embed: [batch,len,emb] 
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), vocab_size, embed_size, name='share_word_embedding', pretrain_embedding=conf.pretrain_emb)
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_mask_turn = tf.tile(tf.expand_dims(s2_mask, 1), [1, s1_turn, 1])  # [batch,turn,len]
        s2_mask_norm = tf.reshape(s2_mask_turn, [-1, s2_length])  # [batch*turn,len]
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, -1), tf.int32)  # [batch]

        # ====== s2 char embedding
        # char_s2_embed: [batch,len,4,emb]
        char_s2_embed, _ = embedding(tf.expand_dims(self.char_s2, -1), cvocab_size, cembed_size, name='share_char_embedding')
        char_s2_mask = mask_nonpad_from_embedding(char_s2_embed)  # [batch,len,4] 1 for nonpad; 0 for pad
        char_s2_embed = tf.reshape(char_s2_embed, [-1, char_len, cembed_size])  # [batch*len,4,emb]

        # 方式1:char cnn + max-pooling
        # char_s2_conv = conv_multi_kernel(char_s2_embed, char_filter, kernels_size=kernels_size,
        #                                  padding='same', use_bias=True, activation=tf.nn.relu,
        #                                  name="char_conv", layer_norm_finally=False)  # [batch*len,4,filter*3]
        # # max-pooling
        # # remove_pad 将pad的vector变为极小
        # char_s2_conv += tf.expand_dims(tf.reshape(1. - char_s2_mask, [-1, char_len]), -1) * -1e9  # [batch*len,4,1]
        # char_s2_conv = tf.reduce_max(char_s2_conv, axis=1)  # [batch*len,hid]
        # char_s2_conv = tf.reshape(char_s2_conv, [-1, s2_length, char_filter * len(kernels_size)])  # [batch,len,hid]
        # char_s2_conv = tf.layers.dropout(char_s2_conv, self.dropout_rate)

        # 方式2:char mean-pooling
        char_s2_sum = tf.reduce_sum(char_s2_embed, axis=-2)  # [batch*len,hid]
        char_s2_num = tf.reduce_sum(tf.reshape(char_s2_mask, [-1, char_len]), -1, keepdims=True)  # [batch*len,1]
        char_s2_mean = char_s2_sum / char_s2_num  # # [batch*len,hid]

        char_s2_mean = tf.reshape(char_s2_mean, [-1, s2_length, cembed_size])  # [batch,len,hid]
        # char_s2_mean = tf.layers.dropout(char_s2_mean, self.dropout_rate)

        # ====== fuse word and char embedding for s2
        s2_embed += char_s2_mean  # seem not convergence -_-
        # s2_embed = tf.concat([s2_embed, char_s2_mean], axis=-1)  # [batch,len,emb+]
        # [batch,len,300]
        # s2_embed = tf.layers.dense(s2_embed, 300, use_bias=False, kernel_initializer=tf.contrib.layers.xavier_initializer(), name='word_fuse', reuse=tf.AUTO_REUSE)
        s2_embed_turn = tf.tile(tf.expand_dims(s2_embed, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,300]
        s2_embed_norm = tf.reshape(s2_embed_turn, [-1, s2_length, 300])  # [batch*turn,len,300]

        # ====== representation and interaction

        # ====== word representation
        if use_word:
            word_interaction = interaction_matching_batch(multi_s1_embed_norm,
                                                          s2_embed_norm,
                                                          embed_size,
                                                          multi_s1_mask_norm,
                                                          s1_turn,
                                                          scope='word_interaction_matching')
            loss_word = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=word_interaction))
            self.loss_lst.append(loss_word)
            self.y_prob_lst.append(tf.nn.softmax(word_interaction))

        # ====== rnn representation
        rnn_dim = 300
        if use_seq:
            with tf.variable_scope("rnn_embeddings"):
                sentence_gru_cell = tf.contrib.rnn.GRUCell(rnn_dim)
                with tf.variable_scope('sentence_gru'):
                    # [batch*turn,len,hid]
                    s1_gru_hidden_norm, _ = tf.nn.dynamic_rnn(sentence_gru_cell, multi_s1_embed_norm, sequence_length=multi_s1_seqlen_norm, dtype=tf.float32)

                with tf.variable_scope('sentence_gru', reuse=True):
                    # [batch,len,hid]
                    s2_gru_hidden, _ = tf.nn.dynamic_rnn(sentence_gru_cell, s2_embed, sequence_length=s2_seqlen, dtype=tf.float32)

            s2_gru_hidden_turn = tf.tile(tf.expand_dims(s2_gru_hidden, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,hid]
            s2_gru_hidden_norm = tf.reshape(s2_gru_hidden_turn, [-1, s2_length, rnn_dim])  # [batch*turn,len,hid]

            seq_interaction = interaction_matching_batch(s1_gru_hidden_norm,
                                                         s2_gru_hidden_norm,
                                                         rnn_dim,
                                                         multi_s1_mask_norm,
                                                         s1_turn,
                                                         scope='seg_interaction_matching')
            loss_seq = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=seq_interaction))
            self.loss_lst.append(loss_seq)
            self.y_prob_lst.append(tf.nn.softmax(seq_interaction))

        # ====== conv representation
        if use_conv:
            conv_dim = 50
            kernels = [1, 2, 3, 4]
            with tf.variable_scope("conv_embeddings"):
                s1_conv_hidden_norm = conv_multi_kernel(multi_s1_embed_norm, conv_dim, kernels_size=kernels,
                                                        padding='same', use_bias=True, activation=tf.nn.relu,
                                                        name="seq_conv", layer_norm_finally=True)  # [batch*turn,len,filter*4]

                s2_conv_hidden = conv_multi_kernel(s2_embed, conv_dim, kernels_size=kernels,
                                                   padding='same', use_bias=True, activation=tf.nn.relu,
                                                   name="seq_conv", layer_norm_finally=True)  # [batch,len,filter*4]

                s2_conv_hidden_turn = tf.tile(tf.expand_dims(s2_conv_hidden, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,filter*4]
                s2_conv_hidden_norm = tf.reshape(s2_conv_hidden_turn, [-1, s2_length, conv_dim * len(kernels)])  # [batch*turn,len,filter*4]

            conv_interaction = interaction_matching_batch(s1_conv_hidden_norm,
                                                          s2_conv_hidden_norm,
                                                          conv_dim * len(kernels),
                                                          multi_s1_mask_norm,
                                                          s1_turn,
                                                          scope='conv_interaction_matching')
            loss_conv = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=conv_interaction))
            self.loss_lst.append(loss_conv)
            self.y_prob_lst.append(tf.nn.softmax(conv_interaction))

        # ====== self attention
        if use_self:
            with tf.variable_scope("self_att_embeddings"):
                s1_self_att_hidden_norm = multihead_attention(multi_s1_embed_norm,
                                                              multi_s1_embed_norm,
                                                              multi_s1_mask_norm,
                                                              embed_size,
                                                              conf.num_heads,
                                                              self.dropout_rate)

                s2_self_att_hidden = multihead_attention(s2_embed,
                                                         s2_embed,
                                                         s2_mask,
                                                         embed_size,
                                                         conf.num_heads,
                                                         self.dropout_rate)

                s2_self_att_hidden_turn = tf.tile(tf.expand_dims(s2_self_att_hidden, 1), [1, s1_turn, 1, 1])  # [batch,turn,len,hid]
                s2_self_att_hidden_norm = tf.reshape(s2_self_att_hidden_turn, [-1, s2_length, embed_size])  # [batch*turn,len,hid]

            self_att_interaction = interaction_matching_batch(s1_self_att_hidden_norm,
                                                              s2_self_att_hidden_norm,
                                                              embed_size,
                                                              multi_s1_mask_norm,
                                                              s1_turn,
                                                              scope='self_att_interaction_matching')
            loss_self = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=self_att_interaction))
            self.loss_lst.append(loss_self)
            self.y_prob_lst.append(tf.nn.softmax(self_att_interaction))

        # cross attention
        if use_cross:
            with tf.variable_scope("cross_att_embeddings"):
                s1_cross_att_hidden_norm = multihead_attention(multi_s1_embed_norm,
                                                               s2_embed_norm,
                                                               s2_mask_norm,
                                                               embed_size,
                                                               conf.num_heads,
                                                               self.dropout_rate)

                s2_cross_attn_hidden_norm = multihead_attention(s2_embed_norm,
                                                                multi_s1_embed_norm,
                                                                multi_s1_mask_norm,
                                                                embed_size,
                                                                conf.num_heads,
                                                                self.dropout_rate)

            cross_att_interaction = interaction_matching_batch(s1_cross_att_hidden_norm,
                                                               s2_cross_attn_hidden_norm,
                                                               embed_size,
                                                               multi_s1_mask_norm,
                                                               s1_turn,
                                                               scope='cross_att_interaction_matching')
            loss_cross = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=self.target, logits=cross_att_interaction))
            self.loss_lst.append(loss_cross)
            self.y_prob_lst.append(tf.nn.softmax(cross_att_interaction))

        self.loss = sum(self.loss_lst)
        y_prob = sum(self.y_prob_lst)  # [batch,2]
        y_prob = y_prob[:, 1]
        self.y_prob = tf.identity(y_prob, 'y_prob')

        with tf.name_scope("accuracy"):
            self.correct = tf.equal(
                tf.cast(tf.greater_equal(self.y_prob, 2.5), tf.int32),  # 因为加了5个0-1的概率，随意中值用2.5
                self.target)
            self.accuracy = tf.reduce_mean(tf.cast(self.correct, tf.float32))

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    @classmethod
    def sent2ids(cls, sent, word2id, max_word_len=None):
        # sent: 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        words = sent.split(' ')
        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
        if max_word_len:
            token_ids = token_ids[:max_word_len - 1]
        token_ids.append(word2id['<eos>'])
        # token_ids = pad_sequences([token_ids], padding='post', maxlen=max_word_len)[0]
        return token_ids  # [len]

    @classmethod
    def multi_sent2ids(cls, multi_sent, word2id, max_word_len=None, max_turn=None):
        # multi_sent: sent的list
        # sent: 已分好词 ' '隔开
        # 不能在形成batch时才动态补齐长度和轮数，因为tfrecord存储会平展为一维，并没有足够信息恢复为不定的[turn,len]
        # 至少需要固定一个(turn或len),后面才能恢复,且需要补齐为矩阵turn * len,后面才可reshape恢复
        # 这里固定turn
        multi_token_ids = []
        for sent in multi_sent:
            multi_token_ids.append(cls.sent2ids(sent, word2id, max_word_len=max_word_len))
        if max_turn is None:  # turn需固定来补齐句子长度
            max_turn = conf.max_turn
        multi_token_ids = multi_token_ids[-max_turn:]  # 若大于截取后n轮
        multi_token_ids = multi_token_ids + [[]] * (max_turn - len(multi_token_ids))  # 若不足则后面补齐
        multi_token_ids = utils.pad_sequences(multi_token_ids, padding='post', value=0)  # 补齐每个sent长度
        return multi_token_ids  # [turn(5), len]

    @classmethod
    def sent2ids_char(cls, sent, word2id, char2id, max_word_len=None, max_char=None, pad_eos=False):
        # sent: 已分好词 ' '隔开
        # char固定,形成batch时才动态补齐len
        words = sent.split(' ')

        if max_word_len:
            if pad_eos:
                words = words[:max_word_len - 1]  # 留末尾放<eos>
            else:
                words = words[:max_word_len]  # 留末尾放<eos>

        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]  # [len]

        if max_char is None:
            max_char = conf.max_char
        char_token_ids = []  # [len,char]
        for word in words:
            char_ids = [char2id.get(char, char2id['<unk>']) for char in word[-max_char:]]  # 超长优先截取后面
            char_ids += [char2id['<pad>']] * (max_char - len(char_ids))  # 不足后面补齐0
            char_token_ids.append(char_ids)

        if pad_eos:
            # 末位补<eos>
            token_ids.append(word2id['<eos>'])
            char_token_ids.append([char2id['<pad>']] * max_char)  # <eos>的char用char_pad补齐

        assert len(token_ids) == len(char_token_ids)
        return token_ids, char_token_ids  # [len]  [len,char]

    @classmethod
    def multi_sent2ids_char(cls, multi_sent, word2id, char2id, max_word_len=None, max_turn=None, max_char=None):
        # 返回char
        # multi_sent: sent的list
        # sent: 已分好词 ' '隔开
        # 不能在形成batch时才动态补齐所有turn,len,char，因为tfrecord存储会平展为一维，并没有足够信息恢复为不定的[turn,len,char]
        # 至少需要固定turn和char,补齐为矩阵turn*len*char,后面才可reshape恢复
        # 这里固定turn/char 让len动态
        if max_char is None:  # 固定且补齐char
            max_char = conf.max_char
        if max_turn is None:  # 固定且补齐turn
            max_turn = conf.max_turn
        multi_token_ids = []
        multi_char_token_ids = []
        for sent in multi_sent:
            token_ids, char_token_ids = cls.sent2ids_char(sent, word2id, char2id, max_word_len=max_word_len, max_char=max_char)
            multi_token_ids.append(token_ids)  # [turn,len]
            multi_char_token_ids.append(char_token_ids)  # [turn,len,char]

        # 补齐turn
        multi_token_ids = multi_token_ids[-max_turn:]  # 若大于截取后n轮
        multi_token_ids = multi_token_ids + [[]] * (max_turn - len(multi_token_ids))  # 若不足则后面补齐
        multi_token_ids = utils.pad_sequences(multi_token_ids, padding='post', value=0)  # 补齐每个sent长度

        multi_char_token_ids = multi_char_token_ids[-max_turn:]  # 若大于截取后n轮
        multi_char_token_ids = multi_char_token_ids + [[]] * (max_turn - len(multi_char_token_ids))  # 若不足则后面补齐
        multi_char_token_ids = utils.pad_sequences(multi_char_token_ids, padding='post', value=0)  # 补齐每个sent长度

        return multi_token_ids, multi_char_token_ids  # [turn(5),len] [turn,len,char(4)]

    def create_feed_dict_from_data(self, data, ids, mode='train'):
        # data:数据已经转为id, data不同字段保存该段字段全量数据
        batch_multi_s1 = [data['multi_s1'][i] for i in ids]
        batch_s2 = [data['s2'][i] for i in ids]
        batch_target = [data['target'][i] for i in ids]
        batch_char_multi_s1 = [data['char_multi_s1'][i] for i in ids]
        batch_char_s2 = [data['char_s2'][i] for i in ids]

        # 多轮必定需要补齐，就不做判断了
        batch_multi_s1 = utils.pad_sequences(batch_multi_s1, padding='post')
        batch_char_multi_s1 = utils.pad_sequences(batch_char_multi_s1, padding='post')
        batch_char_s2 = utils.pad_sequences(batch_char_s2, padding='post')

        if len(set([len(e) for e in batch_s2])) != 1:  # 长度不等
            batch_s2 = utils.pad_sequences(batch_s2, padding='post')
        feed_dict = {
            self.multi_s1: batch_multi_s1,
            self.s2: batch_s2,
            self.target: batch_target,
            self.char_multi_s1: batch_char_multi_s1,
            self.char_s2: batch_char_s2,
        }
        if mode == 'train': feed_dict['num'] = len(batch_s2)
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_features(self, features, mode='train'):
        # feature:tfrecord数据的example, 每个features的不同字段包括该字段一个batch数据
        feed_dict = {
            self.multi_s1: features['multi_s1'],
            self.s2: features['s2'],
            self.target: features['target'],
            self.char_multi_s1: features['char_multi_s1'],
            self.char_s2: features['char_s2'],
        }
        if mode == 'train': feed_dict['num'] = len(features['s2'])
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_raw(self, batch_multi_s1, batch_s2, batch_target, token2id_dct, mode='infer'):
        word2id = token2id_dct['word2id']
        char2id = token2id_dct['char2id']

        feed_multi_s1 = []
        feed_char_multi_s1 = []
        for multi_s1 in batch_multi_s1:
            sent_lst = multi_s1.split('$$$')
            multi_token_ids, multi_char_token_ids = self.multi_sent2ids_char(sent_lst, word2id, char2id)
            feed_multi_s1.append(multi_token_ids)
            feed_char_multi_s1.append(multi_char_token_ids)

        feed_s2 = []
        feed_char_s2 = []
        for s2 in batch_s2:
            token_ids, char_token_ids = self.sent2ids_char(s2, word2id, char2id)
            feed_s2.append(token_ids)
            feed_char_s2.append(char_token_ids)

        feed_dict = {
            self.multi_s1: utils.pad_sequences(feed_multi_s1, padding='post'),
            self.char_multi_s1: utils.pad_sequences(feed_char_multi_s1, padding='post'),
            self.s2: utils.pad_sequences(feed_s2, padding='post'),
            self.char_s2: utils.pad_sequences(feed_char_s2, padding='post'),
        }

        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        if mode == 'infer':
            return feed_dict

        if mode in ['train', 'dev']:
            assert batch_target, 'batch_target should not be None when mode is train or dev'
            feed_dict[self.target] = batch_target
            return feed_dict

        raise ValueError(f'mode type {mode} not support')

    @classmethod
    def generate_data(cls, file, token2id_dct):
        word2id = token2id_dct['word2id']
        char2id = token2id_dct['char2id']
        data = {
            'multi_s1': [],
            's2': [],
            'target': [],
            'char_multi_s1': [],
            'char_s2': [],
        }
        with open(file, 'r', encoding='U8') as f:
            for i, line in enumerate(f):
                item = line.strip().split('\t')
                if len(item) != 3:
                    print('error item:', repr(line))
                    continue
                multi_s1 = item[0].split('$$$')
                s2 = item[1]
                target = item[2]
                multi_s1_ids, char_multi_s1_ids = cls.multi_sent2ids_char(multi_s1, word2id, char2id, max_word_len=50)
                s2_ids, char_s2_ids = cls.sent2ids_char(s2, word2id, char2id, max_word_len=50)
                target_id = int(target)
                if i < 5:  # check
                    print(f'check {i}:')
                    print(f'{multi_s1} -> {multi_s1_ids}')
                    print(f'{s2} -> {s2_ids}')
                    print(f'{target} -> {target_id}')
                    print(f'{multi_s1} -> {char_multi_s1_ids}')
                    print(f'{s2} -> {char_s2_ids}')
                data['multi_s1'].append(multi_s1_ids)
                data['s2'].append(s2_ids)
                data['target'].append(target_id)
                data['char_multi_s1'].append(char_multi_s1_ids)
                data['char_s2'].append(char_s2_ids)
        data['num_data'] = len(data['s2'])
        return data

    @classmethod
    def generate_tfrecord(cls, file, token2id_dct, tfrecord_file):
        from qiznlp.common.tfrecord_utils import items2tfrecord
        word2id = token2id_dct['word2id']
        char2id = token2id_dct['char2id']

        def items_gen():
            with open(file, 'r', encoding='U8') as f:
                for i, line in enumerate(f):
                    item = line.strip().split('\t')
                    if len(item) != 3:
                        print('error item:', repr(line))
                        continue
                    try:
                        multi_s1 = item[0].split('$$$')
                        s2 = item[1]
                        target = item[2]
                        multi_s1_ids, char_multi_s1_ids = cls.multi_sent2ids_char(multi_s1, word2id, char2id, max_word_len=50)
                        s2_ids, char_s2_ids = cls.sent2ids_char(s2, word2id, char2id, max_word_len=50)
                        target_id = int(target)
                        if i < 5:  # check
                            print(f'check {i}:')
                            print(f'{multi_s1} -> {multi_s1_ids}')
                            print(f'{s2} -> {s2_ids}')
                            print(f'{target} -> {target_id}')
                            print(f'{multi_s1} -> {char_multi_s1_ids}')
                            print(f'{s2} -> {char_s2_ids}')
                        d = {
                            'multi_s1': multi_s1_ids,
                            's2': s2_ids,
                            'target': target_id,
                            'char_multi_s1': char_multi_s1_ids,
                            'char_s2': char_s2_ids,
                        }
                        yield d
                    except Exception as e:
                        print('Exception occur in items_gen()!\n', e)
                        continue

        count = items2tfrecord(items_gen(), tfrecord_file)
        return count

    @classmethod
    def load_tfrecord(cls, tfrecord_file, batch_size=128, index=None, shard=None):
        from qiznlp.common.tfrecord_utils import tfrecord2dataset
        if not os.path.exists(tfrecord_file):
            return None, None
        feat_dct = {
            'multi_s1': tf.VarLenFeature(tf.int64),  # [turn,len]
            's2': tf.VarLenFeature(tf.int64),  # [len]
            'target': tf.FixedLenFeature([], tf.int64),
            'char_multi_s1': tf.VarLenFeature(tf.int64),  # [turn,len,char]
            'char_s2': tf.VarLenFeature(tf.int64),  # [len,char]
        }
        shape_dct = {
            'multi_s1': [conf.max_turn, -1],
            'char_multi_s1': [conf.max_turn, -1, conf.max_char],
            'char_s2': [-1, conf.max_char],
        }
        dataset, count = tfrecord2dataset(tfrecord_file, feat_dct, shape_dct=shape_dct, batch_size=batch_size, auto_pad=True, index=index, shard=shard)
        return dataset, count

    def get_signature_export_model(self):
        inputs_dct = {
            'multi_s1': self.multi_s1,
            's2': self.s2,
            'dropout_rate': self.dropout_rate,
            'char_multi_s1': self.char_multi_s1,
            'char_s2': self.char_s2,

        }
        outputs_dct = {
            'y_prob': self.y_prob,
        }
        return inputs_dct, outputs_dct

    @classmethod
    def get_signature_load_pbmodel(cls):
        inputs_lst = ['multi_s1', 's2', 'dropout_rate', 'char_multi_s1', 'char_s2']
        outputs_lst = ['y_prob']
        return inputs_lst, outputs_lst

    @classmethod
    def from_pbmodel(cls, pbmodel_dir, sess):
        meta_graph_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], pbmodel_dir)  # 从pb模型载入graph和variable,绑定到sess
        signature = meta_graph_def.signature_def  # 取出signature_def
        default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
        inputs_lst, output_lst = cls.get_signature_load_pbmodel()  # 类方法
        pb_dict = {}
        for k in inputs_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].inputs[k].name)  # 从signature中获取输入输出的tensor_name,并从graph中取出
        for k in output_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].outputs[k].name)
        model = cls(build_graph=False)  # 里面不再构造图
        for k, v in pb_dict.items():
            setattr(model, k, v)  # 绑定必要的输入输出到实例
        return model

    @classmethod
    def from_ckpt_meta(cls, ckpt_name, sess, graph):
        with graph.as_default():
            saver = tf.train.import_meta_graph(ckpt_name + '.meta', clear_devices=True)
            sess.run(tf.global_variables_initializer())

        model = cls(build_graph=False)  # 里面不再构造图
        # 绑定必要的输入输出到实例
        model.multi_s1 = graph.get_tensor_by_name('multi_s1:0')
        model.s2 = graph.get_tensor_by_name('s2:0')
        model.dropout_rate = graph.get_tensor_by_name('dropout_rate:0')
        model.char_multi_s1 = graph.get_tensor_by_name('char_multi_s1:0')
        model.char_s2 = graph.get_tensor_by_name('char_s2:0')
        model.y_prob = graph.get_tensor_by_name('y_prob:0')

        saver.restore(sess, ckpt_name)
        print(f':: restore success! {ckpt_name}')
        return model, saver



#!/usr/bin/env python
# coding=utf-8
import jieba
import numpy as np
import tensorflow as tf
import os, time, re, sys

""" 
deploy cls_model example(for toutiao dataset)
"""
curr_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(curr_dir)
sys.path.append(curr_dir + '/..')

import qiznlp.common.utils as utils
from qiznlp.model.cls_model import Model


class Deplpy_CLS_Model():
    def __init__(self, ckpt_name=None, pbmodel_dir=None):
        assert ckpt_name or pbmodel_dir, 'ues at least one way'
        self.graph = tf.Graph()
        self.config = tf.ConfigProto(allow_soft_placement=True,
                                     gpu_options=tf.GPUOptions(allow_growth=True),
                                     )
        self.sess = tf.Session(config=self.config, graph=self.graph)

        self.token2id_dct = {
            'word2id': utils.Any2Id.from_file(f'{curr_dir}/../data/toutiao_cls_word2id.dct', use_line_no=True),
            'label2id': utils.Any2Id.from_file(f'{curr_dir}/../data/toutiao_cls_label2id.dct', use_line_no=True),
        }
        self.jieba = jieba.Tokenizer()
        self.tokenize = lambda t: self.jieba.lcut(re.sub(r'\s+', '，', t))
        self.cut = lambda t: ' '.join(self.tokenize(t))
        if ckpt_name:
            self.load_from_ckpt_meta(ckpt_name)
        else:
            self.load_from_pbmodel(pbmodel_dir)

        self.id2label = self.token2id_dct['label2id'].get_reverse()

    def load_from_ckpt_meta(self, ckpt_name):
        self.model, self.saver = Model.from_ckpt_meta(ckpt_name, self.sess, self.graph)

    def load_from_pbmodel(self, pbmodel_dir):
        self.model = Model.from_pbmodel(pbmodel_dir, self.sess)

    def predict(self, sent, need_cut=True):
        if need_cut:
            sent = self.cut(sent)
        feed_dict = self.model.create_feed_dict_from_raw([sent], [], self.token2id_dct)
        prob = self.sess.run(self.model.y_prob, feed_dict)[0]
        pred = np.argmax(prob)  # [batch]
        return self.id2label[pred]


if __name__ == '__main__':
    os.environ['CUDA_VISIBLE_DEVICES'] = '0'  # 使用CPU设为'-1'

    # option 1. from ckpt
    # assume ckpt file is 'cls_ckpt_toutiao1/trans_mhattnpool-7-0.58-0.999.ckpt-854'
    dcm = Deplpy_CLS_Model(ckpt_name=f'{curr_dir}/../run/cls_ckpt_toutiao1/trans_mhattnpool-7-0.58-0.999.ckpt-854')
    print(dcm.predict('去日本的邮轮游需要5万的资产证明吗？'))


    # option 2. from pbmodel
    # firstly export pbmodel in [cls_pbmodel_dir] with:
    """
    rm_cls = Run_Model_Cls('trans_mhattnpool')
    rm.export_model(f'{curr_dir}/../run/cls_pbmodel_dir')
    """
    # then exec follow:
    dcm = Deplpy_CLS_Model(pbmodel_dir=f'{curr_dir}/../run/cls_pbmodel_dir')
    print(dcm.predict('去日本的邮轮游需要5万的资产证明吗？'))



#!/usr/bin/env python
# encoding=utf-8

import os
import shutil

def qiznlp_init():
    cwd = os.getcwd()
    curr_dir = os.path.dirname(__file__)

    print('copying ...')

    shutil.copytree(f'{curr_dir}/model', f'{cwd}/model')
    print('copy model_dir finish')

    shutil.copytree(f'{curr_dir}/run', f'{cwd}/run')
    os.remove(f'{cwd}/run/run_base.py')
    print('copy run_dir finish')

    shutil.copytree(f'{curr_dir}/deploy', f'{cwd}/deploy')
    print('copy deploy_dir finish')

    shutil.copytree(f'{curr_dir}/data', f'{cwd}/data')
    print('copy data_dir finish')

    shutil.copytree(f'{curr_dir}/common/modules/bert/chinese_L-12_H-768_A-12',
                    f'{cwd}/common/modules/bert/chinese_L-12_H-768_A-12')
    print('copy bert_model_dir finish')






import os
import tensorflow as tf
from tensorflow.python.util import nest
import numpy as np

curr_dir = os.path.dirname(os.path.realpath(__file__))

from qiznlp.common.modules.common_layers import shape_list, mask_nonpad_from_embedding, add_timing_signal_1d, get_timing_signal_1d, shift_right, split_heads
from qiznlp.common.modules.embedding import embedding, proj_logits
from qiznlp.common.modules.encoder import transformer_encoder, transformer_decoder, EncDecAttention
from qiznlp.common.modules.birnn import Bi_RNN, RNN
from qiznlp.common.modules.beam_search import beam_search, greedy_search, get_state_shape_invariants
import qiznlp.common.utils as utils

conf = utils.dict2obj({
    'max_turn': 5,
    'vocab_size': 4000,
    'embed_size': 300,
    'hidden_size': 300,
    'num_heads': 6,
    'num_encoder_layers': 6,
    'num_decoder_layers': 6,
    'dropout_rate': 0.2,
    'lr': 1e-3,
    'pretrain_emb': None,
    'beam_size': 5,
    'max_decode_len': 50,
    'eos_id': 2,
})


class Model(object):
    def __init__(self, build_graph=True, **kwargs):
        self.conf = conf
        self.run_model = kwargs.get('run_model', None)  # acquire outside run_model instance
        if build_graph:
            # build placeholder
            self.build_placeholder()
            # build model
            self.model_name = kwargs.get('model_name', 'HRED')
            {
                'HRED': self.build_model1,  # 1507.04808 Building End-To-End Dialogue Systems Using Generative Hierarchical Neural Network Models https://github.com/julianser/hed-dlg 
                'HRAN': self.build_model2,  # 1701.07149 Hierarchical Recurrent Attention Network for Response Generation https://github.com/chenhongshen/HVMN/tree/master/model
                'RECOSA': self.build_model3,  # 1907.05339 ReCoSa Detecting the Relevant Contexts with Self-Attention for Multi-turn Dialogue Generation https://github.com/zhanghainan/ReCoSa 
                # add new here
            }[self.model_name]()
            print(f'model_name: {self.model_name} build graph ok!')

    def build_placeholder(self):
        # placeholder
        # 原则上模型输入输出不变，不需换新model
        self.multi_s1 = tf.placeholder(tf.int32, [None, None, None], name='multi_s1')  # [batch,turn,len]
        self.s2 = tf.placeholder(tf.int32, [None, None], name='s2')
        self.dropout_rate = tf.placeholder(tf.float32, name='dropout_rate')

    def build_model1(self):
        self.uttn_enc_hidden_size = 256
        self.ctx_enc_hidden_size = 256
        batch_size, num_turns, length = shape_list(self.multi_s1)

        # embedding
        # [batch,len,turn,embed]
        multi_s1_embed, _ = embedding(tf.expand_dims(self.multi_s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        # [batch,len,embed]
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)

        # uttn encoder
        uttn_input = tf.reshape(multi_s1_embed, [-1, length, conf.embed_size])  # [batch*turn,len,embed]
        uttn_mask = mask_nonpad_from_embedding(uttn_input)  # [batch*turn,len] 1 for nonpad; 0 for pad
        uttn_seqlen = tf.cast(tf.reduce_sum(uttn_mask, axis=-1), tf.int32)  # [batch*turn]
        # uttn-gru
        self.encoder_uttn_rnn = RNN(cell_name='GRUCell', name='uttn_enc', hidden_size=self.uttn_enc_hidden_size, dropout_rate=self.dropout_rate)
        _, uttn_embed = self.encoder_uttn_rnn(uttn_input, uttn_seqlen)  # [batch*turn,hid]
        uttn_embed = tf.reshape(uttn_embed, [batch_size, num_turns, self.uttn_enc_hidden_size])  # [batch,turn,hid]  # 之后turn相当于len

        # ctx encoder
        ctx_mask = mask_nonpad_from_embedding(uttn_embed)  # [batch,turn] 1 for nonpad; 0 for pad
        ctx_seqlen = tf.cast(tf.reduce_sum(ctx_mask, axis=-1), tf.int32)  # [batch]
        # ctx-gru
        self.encoder_ctx_rnn = RNN(cell_name='GRUCell', name='ctx_enc', hidden_size=self.ctx_enc_hidden_size, dropout_rate=self.dropout_rate)
        _, ctx_embed = self.encoder_ctx_rnn(uttn_embed, ctx_seqlen)  # [batch,hid]

        # rnn decoder train (no attention)
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, axis=-1), tf.int32)  # [batch]

        decoder_input = shift_right(s2_embed)  # 用pad当做eos
        decoder_input = tf.layers.dropout(decoder_input, rate=self.dropout_rate)  # dropout

        # 输入拼上ctx
        decoder_ctx = tf.tile(tf.expand_dims(ctx_embed, axis=1), [1, shape_list(decoder_input)[1], 1])  # [batch,len,hid]
        decoder_input = tf.concat([decoder_input, decoder_ctx], axis=2)

        self.decoder_rnn = RNN(cell_name='GRUCell', name='dec', hidden_size=conf.embed_size, dropout_rate=self.dropout_rate)
        decoder_output, decoder_state = self.decoder_rnn(decoder_input, s2_seqlen)

        logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')

        onehot_s2 = tf.one_hot(self.s2, depth=conf.vocab_size)  # [batch,len,vocab]

        xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
        weights = tf.to_float(tf.not_equal(self.s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

        loss_num = xentropy * weights  # [batch,len]
        loss_den = weights  # [batch,len]

        loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar
        self.loss = loss

        # rnn decoder infer (no attention)
        # 放在cache里面的在后面symbols_to_logits_fn函数中都会变成batch * beam
        cache = {'state': self.decoder_rnn.cell.zero_state(batch_size, tf.float32),  # [batch,hid]
                 'ctx': ctx_embed,  # [batch,hid]
                 }

        def symbols_to_logits_fn(ids, i, cache):
            # ids [batch,length]
            pred_target = ids[:, -1:]  # [batch,1] 截取最后一个
            target_embed, _ = embedding(tf.expand_dims(pred_target, axis=-1), conf.vocab_size, conf.embed_size, 'share_embedding')  # [batch,1,embed]
            decoder_input = tf.squeeze(target_embed, axis=1)  # [batch,embed]

            # 输入加上ctx
            decoder_input = tf.concat([decoder_input, cache['ctx']], axis=-1)

            # run rnn
            decoder_output, cache['state'] = self.decoder_rnn.one_step(decoder_input, cache['state'])

            logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')

            return logits, cache

        initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

        def greedy_search_wrapper():
            """ Greedy Search """
            decoded_ids, scores = greedy_search(
                symbols_to_logits_fn,
                initial_ids,
                conf.max_decode_len,
                cache=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        def beam_search_wrapper():
            """ Beam Search """
            decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
                symbols_to_logits_fn,
                initial_ids,
                conf.beam_size,
                conf.max_decode_len,
                conf.vocab_size,
                alpha=0,
                states=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        decoded_ids, scores = tf.cond(tf.equal(conf.beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

        self.decoded_ids = tf.identity(decoded_ids, name='decoded_ids')  # [batch,beam/1,len]
        self.scores = tf.identity(scores, name='scores')  # [batch,beam/1]

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model2(self):
        self.uttn_enc_hidden_size = 256
        self.ctx_enc_hidden_size = 256
        batch_size, num_turns, length = shape_list(self.multi_s1)

        # embedding
        # [batch,len,turn,embed]
        multi_s1_embed, _ = embedding(tf.expand_dims(self.multi_s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        # [batch,len,embed]
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)

        # uttn encoder
        uttn_input = tf.reshape(multi_s1_embed, [-1, length, conf.embed_size])  # [batch*turn,len,embed]
        uttn_mask = mask_nonpad_from_embedding(uttn_input)  # [batch*turn,len] 1 for nonpad; 0 for pad
        uttn_seqlen = tf.cast(tf.reduce_sum(uttn_mask, axis=-1), tf.int32)  # [batch*turn]
        # uttn-gru
        self.encoder_uttn_rnn = Bi_RNN(cell_name='GRUCell', name='uttn_enc', hidden_size=self.uttn_enc_hidden_size, dropout_rate=self.dropout_rate)
        uttn_repre, uttn_embed = self.encoder_uttn_rnn(uttn_input, uttn_seqlen)  # [batch*turn,len,2hid] [batch*turn,2hid]
        uttn_embed = tf.reshape(uttn_embed, [batch_size, num_turns, self.uttn_enc_hidden_size * 2])  # [batch,turn,2hid]
        uttn_repre = tf.reshape(uttn_repre, [batch_size, num_turns, length, self.uttn_enc_hidden_size * 2])  # [batch,turn,len,2hid]

        # ctx encoder
        ctx_mask = mask_nonpad_from_embedding(uttn_embed)  # [batch,turn] 1 for nonpad; 0 for pad
        ctx_seqlen = tf.cast(tf.reduce_sum(ctx_mask, axis=-1), tf.int32)  # [batch]

        # reverse turn
        uttn_repre = tf.reverse_sequence(uttn_repre, seq_lengths=ctx_seqlen, seq_axis=1, batch_axis=0)

        # ctx-gru
        self.encoder_ctx_rnn = RNN(cell_name='GRUCell', name='ctx_enc', hidden_size=self.ctx_enc_hidden_size, dropout_rate=self.dropout_rate)
        init_ctx_encoder_state = self.encoder_ctx_rnn.cell.zero_state(batch_size, tf.float32)

        # rnn decoder train
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, axis=-1), tf.int32)  # [batch]

        decoder_input = shift_right(s2_embed)  # 用pad当做eos
        decoder_input = tf.layers.dropout(decoder_input, rate=self.dropout_rate)  # dropout

        self.decoder_rnn = RNN(cell_name='GRUCell', name='dec', hidden_size=conf.embed_size, dropout_rate=self.dropout_rate)
        init_decoder_state = self.decoder_rnn.cell.zero_state(batch_size, tf.float32)

        # 两重循环 -_-
        # 外层循环是decoder解码的每一步 time_step
        def loop_condition(time_step, *_):
            return tf.less(time_step, tf.reduce_max(s2_seqlen))

        def loop_body(time_step, dec_rnn_state, dec_rnn_output):
            # cal decoder ctx
            # word level attention

            # 内层循环是对于解码每一步 ctx_rnn上turn的每一步，最终生成ctx序列 turn_step
            def inner_loop_condition(turn_step, *_):
                return tf.less(turn_step, tf.reduce_max(ctx_seqlen))

            def inner_loop_body(turn_step, ctx_rnn_state, ctx_rnn_output):
                # 根据si, ctx_init_state 递归计算多个turn的句子ctx
                q_antecedent = tf.concat([ctx_rnn_state, dec_rnn_state], axis=-1)  # [batch, h]
                q_antecedent = tf.tile(tf.expand_dims(q_antecedent, 1), [1, length, 1])  # [batch,len,h]

                # 抽取每个batch的第i个turn
                # sent_repre [batch,turn,len,h]
                q_antecedent = tf.concat([uttn_repre[:, turn_step, :, :], q_antecedent], -1)  # [batch,len,h] 
                uttn_mask_in_turn = tf.reshape(uttn_mask, [batch_size, num_turns, length])[:, turn_step, :]  # [batch,len]

                # word-level-attn
                h = tf.layers.dense(q_antecedent, 128, activation=tf.nn.tanh, use_bias=True, name='word_level_attn/layer1')
                energy = tf.layers.dense(h, 1, use_bias=True, name='word_level_attn/layer2')  # [batch,len,1]
                energy = tf.squeeze(energy, -1) + (1. - uttn_mask_in_turn) * -1e9
                alpha = tf.nn.softmax(energy)  # [batch,len]
                r_in_turn = tf.reduce_sum(tf.expand_dims(alpha, -1) * uttn_repre[:, turn_step, :, :], 1)  # [batch,h]

                ctx_rnn_output_, ctx_rnn_state = self.encoder_ctx_rnn.one_step(r_in_turn, ctx_rnn_state)

                # attch
                ctx_rnn_output = tf.concat([ctx_rnn_output, tf.expand_dims(ctx_rnn_output_, 1)], 1)  # [batch,turn,h]
                return turn_step + 1, ctx_rnn_state, ctx_rnn_output

            # start inner loop
            final_turn_step, final_state, ctx_rnn_output = tf.while_loop(
                inner_loop_condition,
                inner_loop_body,
                loop_vars=[tf.constant(0, dtype=tf.int32),
                           init_ctx_encoder_state,
                           tf.zeros([batch_size, 0, self.ctx_enc_hidden_size])],
                shape_invariants=[
                    tf.TensorShape([]),
                    nest.map_structure(get_state_shape_invariants, init_ctx_encoder_state),
                    tf.TensorShape([None, None, self.ctx_enc_hidden_size]),
                ])

            # ctx_rnn_output  # [batch,turn,h]
            # dec_rnn_state  # [batch,h]
            # ctx-level-attn
            # q_antecedent = tf.tile(tf.expand_dims(dec_rnn_state, axis=1), [1, num_turns, 1])  # [batch,turn,h]
            # 这样只拿当前batch中的尽可能小的turns数量而不是固定turn
            q_antecedent = tf.tile(tf.expand_dims(dec_rnn_state, axis=1), [1, shape_list(ctx_rnn_output)[1], 1])  # [batch,turn,h]
            q_antecedent = tf.concat([q_antecedent, ctx_rnn_output], 2)  # [batch,turn,h]
            h = tf.layers.dense(q_antecedent, 128, activation=tf.nn.tanh, use_bias=True, name='ctx_level_attn/layer1')
            energy = tf.layers.dense(h, 1, use_bias=True, name='ctx_level_attn/layer2')  # [batch,turn,1]
            energy = tf.squeeze(energy, -1) + (1. - ctx_mask) * -1e9  # [batch,turn]
            alpha = tf.nn.softmax(energy)  # [batch,turn]
            ctx_input_in_dec = tf.reduce_sum(tf.expand_dims(alpha, -1) * ctx_rnn_output, 1)  # [batch,h]

            dec_rnn_input = tf.concat([ctx_input_in_dec, decoder_input[:, time_step, :]], -1)  # [batch,h]
            dec_rnn_output_, dec_rnn_state = self.decoder_rnn.one_step(dec_rnn_input, dec_rnn_state)

            dec_rnn_output = tf.concat([dec_rnn_output, tf.expand_dims(dec_rnn_output_, 1)], 1)

            return time_step + 1, dec_rnn_state, dec_rnn_output

        # start outer loop
        final_time_step, final_state, dec_rnn_output = tf.while_loop(
            loop_condition,
            loop_body,
            loop_vars=[tf.constant(0, dtype=tf.int32),
                       init_decoder_state,
                       tf.zeros([batch_size, 0, conf.embed_size])],
            shape_invariants=[
                tf.TensorShape([]),
                nest.map_structure(get_state_shape_invariants, init_decoder_state),
                tf.TensorShape([None, None, conf.embed_size]),
            ])

        decoder_output = dec_rnn_output

        logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')

        onehot_s2 = tf.one_hot(self.s2, depth=conf.vocab_size)  # [batch,len,vocab]

        xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
        weights = tf.to_float(tf.not_equal(self.s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

        loss_num = xentropy * weights  # [batch,len]
        loss_den = weights  # [batch,len]

        loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar
        self.loss = loss

        # rnn decoder infer
        # 放在cache里面的在后面symbols_to_logits_fn函数中都会变成batch * beam
        cache = {'dec_rnn_state': self.decoder_rnn.cell.zero_state(batch_size, tf.float32),  # [batch,hid]
                 'ctx_rnn_state': self.encoder_ctx_rnn.cell.zero_state(batch_size, tf.float32),  # [batch,hid]
                 'uttn_repre': uttn_repre  # [batch,turn,len,2hid]
                 }

        def symbols_to_logits_fn(ids, i, cache):
            # ids [batch,length]
            pred_target = ids[:, -1:]  # [batch,1] 截取最后一个
            target_embed, _ = embedding(tf.expand_dims(pred_target, axis=-1), conf.vocab_size, conf.embed_size, 'share_embedding')  # [batch,1,embed]
            decoder_input = tf.squeeze(target_embed, axis=1)  # [batch,embed]

            dec_rnn_state = cache['dec_rnn_state']

            with tf.variable_scope('', reuse=tf.AUTO_REUSE):
                # 内层循环是对于解码每一步 ctx_rnn上turn的每一步，最终生成ctx序列 turn_step
                def inner_loop_condition(turn_step, *_):
                    return tf.less(turn_step, tf.reduce_max(ctx_seqlen))

                def inner_loop_body(turn_step, ctx_rnn_state, ctx_rnn_output):
                    # 根据si, ctx_init_state 递归计算多个turn的句子ctx
                    q_antecedent = tf.concat([ctx_rnn_state, dec_rnn_state], axis=-1)  # [batch, h]
                    q_antecedent = tf.tile(tf.expand_dims(q_antecedent, 1), [1, length, 1])  # [batch,len,h]

                    # 抽取每个batch的第i个turn
                    # sent_repre [batch,turn,len,h]
                    q_antecedent = tf.concat([cache['uttn_repre'][:, turn_step, :, :], q_antecedent], -1)  # [batch,len,h] 
                    uttn_mask_in_turn = tf.reshape(uttn_mask, [batch_size, num_turns, length])[:, turn_step, :]  # [batch,len]

                    # word-level-attn
                    h = tf.layers.dense(q_antecedent, 128, activation=tf.nn.tanh, use_bias=True, name='word_level_attn/layer1')
                    energy = tf.layers.dense(h, 1, use_bias=True, name='word_level_attn/layer2')  # [batch,len,1]
                    energy = tf.squeeze(energy, -1) + (1. - uttn_mask_in_turn) * -1e9
                    alpha = tf.nn.softmax(energy)  # [batch,len]
                    r_in_turn = tf.reduce_sum(tf.expand_dims(alpha, -1) * cache['uttn_repre'][:, turn_step, :, :], 1)  # [batch,h]

                    ctx_rnn_output_, ctx_rnn_state = self.encoder_ctx_rnn.one_step(r_in_turn, ctx_rnn_state)
                    # attch
                    ctx_rnn_output = tf.concat([ctx_rnn_output, tf.expand_dims(ctx_rnn_output_, 1)], 1)  # [batch,turn,h]
                    return turn_step + 1, ctx_rnn_state, ctx_rnn_output

                # start inner loop
                final_turn_step, final_state, ctx_rnn_output = tf.while_loop(
                    inner_loop_condition,
                    inner_loop_body,
                    loop_vars=[tf.constant(0, dtype=tf.int32),
                               cache['ctx_rnn_state'],
                               tf.zeros([shape_list(cache['ctx_rnn_state'])[0], 0, self.ctx_enc_hidden_size])],
                    shape_invariants=[
                        tf.TensorShape([]),
                        nest.map_structure(get_state_shape_invariants, init_ctx_encoder_state),
                        tf.TensorShape([None, None, self.ctx_enc_hidden_size]),
                    ])

                # ctx_rnn_output  # [batch,turn,h]
                # dec_rnn_state  # [batch,h]
                # ctx-level-attn
                # q_antecedent = tf.tile(tf.expand_dims(dec_rnn_state, axis=1), [1, num_turns, 1])  # [batch,turn,h]
                # 这样只拿当前batch中的尽可能小的turns数量而不是固定turn
                q_antecedent = tf.tile(tf.expand_dims(dec_rnn_state, axis=1), [1, shape_list(ctx_rnn_output)[1], 1])  # [batch,turn,h]
                q_antecedent = tf.concat([q_antecedent, ctx_rnn_output], 2)  # [batch,turn,h]
                h = tf.layers.dense(q_antecedent, 128, activation=tf.nn.tanh, use_bias=True, name='ctx_level_attn/layer1')
                energy = tf.layers.dense(h, 1, use_bias=True, name='ctx_level_attn/layer2')  # [batch,turn,1]
                energy = tf.squeeze(energy, -1) + (1. - ctx_mask) * -1e9  # [batch,turn]
                alpha = tf.nn.softmax(energy)  # [batch,turn]
                ctx_input_in_dec = tf.reduce_sum(tf.expand_dims(alpha, -1) * ctx_rnn_output, 1)  # [batch,h]

                dec_rnn_input = tf.concat([ctx_input_in_dec, decoder_input], -1)  # [batch,h]
                dec_rnn_output_, dec_rnn_state = self.decoder_rnn.one_step(dec_rnn_input, dec_rnn_state)

                cache['dec_rnn_state'] = dec_rnn_state

                logits = proj_logits(dec_rnn_output_, conf.embed_size, conf.vocab_size, name='share_embedding')

                return logits, cache

        initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

        def greedy_search_wrapper():
            """ Greedy Search """
            decoded_ids, scores = greedy_search(
                symbols_to_logits_fn,
                initial_ids,
                conf.max_decode_len,
                cache=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        def beam_search_wrapper():
            """ Beam Search """
            decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
                symbols_to_logits_fn,
                initial_ids,
                conf.beam_size,
                conf.max_decode_len,
                conf.vocab_size,
                alpha=0,
                states=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        decoded_ids, scores = tf.cond(tf.equal(conf.beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

        self.decoded_ids = tf.identity(decoded_ids, name='decoded_ids')  # [batch,beam/1,len]
        self.scores = tf.identity(scores, name='scores')  # [batch,beam/1]

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)
        
    def build_model3(self):
        assert not conf.hidden_size % 2  # 被2整除
        self.uttn_enc_hidden_size = conf.hidden_size // 2
        batch_size, num_turns, length = shape_list(self.multi_s1)

        # embedding
        # [batch,len,turn,embed]
        multi_s1_embed, _ = embedding(tf.expand_dims(self.multi_s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        # [batch,len,embed]
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)

        # uttn encoder
        uttn_input = tf.reshape(multi_s1_embed, [-1, length, conf.embed_size])  # [batch*turn,len,embed]
        uttn_mask = mask_nonpad_from_embedding(uttn_input)  # [batch*turn,len] 1 for nonpad; 0 for pad
        uttn_seqlen = tf.cast(tf.reduce_sum(uttn_mask, axis=-1), tf.int32)  # [batch*turn]
        # uttn-gru
        self.encoder_uttn_rnn = Bi_RNN(cell_name='GRUCell', name='uttn_enc', hidden_size=self.uttn_enc_hidden_size, dropout_rate=self.dropout_rate)
        _, uttn_embed = self.encoder_uttn_rnn(uttn_input, uttn_seqlen)  # [batch*turn,len,2hid] [batch*turn,2hid]
        uttn_embed = tf.reshape(uttn_embed, [batch_size, num_turns, self.uttn_enc_hidden_size * 2])  # [batch,turn,2hid]  # 之后turn相当于len

        # transformer ctx encoder
        encoder_valid_mask = mask_nonpad_from_embedding(uttn_embed)  # [batch,turn] 1 for nonpad; 0 for pad
        encoder_input = add_timing_signal_1d(uttn_embed)  # add position embedding
        encoder_input = tf.layers.dropout(encoder_input, rate=self.dropout_rate)  # dropout

        encoder_output = transformer_encoder(encoder_input, encoder_valid_mask,
                                             hidden_size=conf.hidden_size,
                                             filter_size=conf.hidden_size * 4,
                                             num_heads=conf.num_heads,
                                             num_encoder_layers=conf.num_encoder_layers,
                                             dropout=self.dropout_rate,
                                             attention_dropout=self.dropout_rate,
                                             relu_dropout=self.dropout_rate,
                                             )

        # transformer decoder
        decoder_input = s2_embed
        decoder_valid_mask = mask_nonpad_from_embedding(decoder_input)  # [batch,len] 1 for nonpad; 0 for pad
        decoder_input = shift_right(decoder_input)  # 用pad当做eos
        decoder_input = add_timing_signal_1d(decoder_input)
        decoder_input = tf.layers.dropout(decoder_input, rate=self.dropout_rate)  # dropout

        decoder_output = transformer_decoder(decoder_input, encoder_output, decoder_valid_mask, encoder_valid_mask,
                                             cache=None,
                                             hidden_size=conf.hidden_size,
                                             filter_size=conf.hidden_size * 4,
                                             num_heads=conf.num_heads,
                                             num_decoder_layers=conf.num_decoder_layers,
                                             dropout=self.dropout_rate,
                                             attention_dropout=self.dropout_rate,
                                             relu_dropout=self.dropout_rate,
                                             )

        logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')

        onehot_s2 = tf.one_hot(self.s2, depth=conf.vocab_size)  # [batch,len,vocab]

        xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
        weights = tf.to_float(tf.not_equal(self.s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

        loss_num = xentropy * weights  # [batch,len]
        loss_den = weights  # [batch,len]

        loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar
        self.loss = loss

        # transformer decoder infer
        # 放在cache里面的在后面symbols_to_logits_fn函数中都会变成batch * beam
        # 初始化缓存
        cache = {
            'layer_%d' % layer: {
                # 用以缓存decoder过程前面已计算的k,v
                'k': split_heads(tf.zeros([batch_size, 0, conf.embed_size]), conf.num_heads),
                'v': split_heads(tf.zeros([batch_size, 0, conf.embed_size]), conf.num_heads)
            } for layer in range(conf.num_decoder_layers)
        }
        for layer in range(conf.num_decoder_layers):
            # 对于decoder每层均需与encoder顶层隐状态计算attention,相应均有特定的k,v可缓存
            layer_name = 'layer_%d' % layer
            with tf.variable_scope('decoder/%s/encdec_attention/multihead_attention' % layer_name):
                k_encdec = tf.layers.dense(encoder_output, conf.embed_size, use_bias=False, name='k', reuse=tf.AUTO_REUSE)
                k_encdec = split_heads(k_encdec, conf.num_heads)
                v_encdec = tf.layers.dense(encoder_output, conf.embed_size, use_bias=False, name='v', reuse=tf.AUTO_REUSE)
                v_encdec = split_heads(v_encdec, conf.num_heads)
            cache[layer_name]['k_encdec'] = k_encdec
            cache[layer_name]['v_encdec'] = v_encdec
        cache['encoder_output'] = encoder_output
        cache['encoder_mask'] = encoder_valid_mask

        # position embedding
        position_embedding = get_timing_signal_1d(conf.max_decode_len, conf.embed_size)  # +eos [1,length+1,embed]

        def symbols_to_logits_fn(ids, i, cache):
            ids = ids[:, -1:]  # [batch,1] 截取最后一个
            target_embed, _ = embedding(tf.expand_dims(ids, axis=-1), conf.vocab_size, conf.embed_size, 'share_embedding', reuse=True)  # [batch,1,hidden]

            decoder_input = target_embed + position_embedding[:, i:i + 1, :]  # [batch,1,hidden]

            encoder_output = cache['encoder_output']
            encoder_mask = cache['encoder_mask']

            with tf.variable_scope('', reuse=tf.AUTO_REUSE):
                decoder_output = transformer_decoder(decoder_input, encoder_output, None, encoder_mask,
                                                     cache=cache,  # 注意infer要cache
                                                     hidden_size=conf.embed_size,
                                                     filter_size=conf.embed_size * 4,
                                                     num_heads=6,
                                                     num_decoder_layers=6,
                                                     )
            logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')  # [batch,1,vocab]
            ret = tf.squeeze(logits, axis=1)  # [batch,vocab]
            return ret, cache

        initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

        def greedy_search_wrapper():
            """ Greedy Search """
            decoded_ids, scores = greedy_search(
                symbols_to_logits_fn,
                initial_ids,
                conf.max_decode_len,
                cache=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        def beam_search_wrapper():
            """ Beam Search """
            decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
                symbols_to_logits_fn,
                initial_ids,
                conf.beam_size,
                conf.max_decode_len,
                conf.vocab_size,
                alpha=0,
                states=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        decoded_ids, scores = tf.cond(tf.equal(conf.beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

        self.decoded_ids = tf.identity(decoded_ids, name='decoded_ids')  # [batch,beam/1,len]
        self.scores = tf.identity(scores, name='scores')  # [batch,beam/1]

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    @classmethod
    def sent2ids(cls, sent, word2id, max_word_len=None):
        # sent: 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        words = sent.split(' ')
        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
        if max_word_len:
            token_ids = token_ids[:max_word_len - 1]
        token_ids.append(word2id['<eos>'])
        # token_ids = pad_sequences([token_ids], padding='post', maxlen=max_word_len)[0]
        return token_ids  # [len]

    @classmethod
    def multi_sent2ids(cls, multi_sent, word2id, max_word_len=None, max_turn=None):
        # multi_sent: sent的list
        # sent: 已分好词 ' '隔开
        # 不能在形成batch时才动态补齐长度和轮数，因为tfrecord存储会平展为一维，并没有足够信息恢复为不定的[turn,len]
        # 至少需要固定一个(turn或len),且需要补齐为矩阵turn * len,后面才可通过reshape恢复
        # 这里固定turn
        multi_token_ids = []
        for sent in multi_sent:
            multi_token_ids.append(cls.sent2ids(sent, word2id, max_word_len=max_word_len))
        if max_turn is None:  # turn需固定来补齐句子长度
            max_turn = conf.max_turn
        multi_token_ids = multi_token_ids[-max_turn:]  # 若大于截取后n轮
        multi_token_ids = multi_token_ids + [[]] * (max_turn - len(multi_token_ids))  # 若不足则后面补齐
        multi_token_ids = utils.pad_sequences(multi_token_ids, padding='post', value=0)  # 补齐每个sent长度
        return multi_token_ids  # [turn(5), len]

    def create_feed_dict_from_data(self, data, ids, mode='train'):
        # data:数据已经转为id, data不同字段保存该段字段全量数据
        batch_multi_s1 = [data['multi_s1'][i] for i in ids]
        batch_s2 = [data['s2'][i] for i in ids]

        # 多轮必定需要补齐，就不做判断了
        batch_multi_s1 = utils.pad_sequences(batch_multi_s1, padding='post')

        if len(set([len(e) for e in batch_s2])) != 1:  # 长度不等
            batch_s2 = utils.pad_sequences(batch_s2, padding='post')
        feed_dict = {
            self.multi_s1: batch_multi_s1,
            self.s2: batch_s2,
        }
        if mode == 'train': feed_dict['num'] = len(batch_s2)
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_features(self, features, mode='train'):
        # feature:tfrecord数据的example, 每个features的不同字段包括该字段一个batch数据
        feed_dict = {
            self.multi_s1: features['multi_s1'],
            self.s2: features['s2'],
        }
        if mode == 'train': feed_dict['num'] = len(features['s2'])
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_raw(self, batch_multi_s1, batch_s2, token2id_dct, mode='infer'):
        word2id = token2id_dct['word2id']

        feed_multi_s1 = [self.multi_sent2ids(multi_s1.split('$$$'), word2id) for multi_s1 in batch_multi_s1]

        feed_dict = {
            self.multi_s1: utils.pad_sequences(feed_multi_s1, padding='post'),
        }
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        if mode == 'infer':
            return feed_dict

        if mode in ['train', 'dev']:
            assert batch_s2, 'batch_s2 should not be None when mode is train or dev'
            feed_s2 = [self.sent2ids(s2, word2id) for s2 in batch_s2]
            feed_dict[self.s2] = utils.pad_sequences(feed_s2, padding='post')
            return feed_dict

        raise ValueError(f'mode type {mode} not support')

    @classmethod
    def generate_data(cls, file, token2id_dct):
        word2id = token2id_dct['word2id']
        data = {
            'multi_s1': [],
            's2': [],
        }
        with open(file, 'r', encoding='U8') as f:
            for i, line in enumerate(f):
                item = line.strip().split('\t')
                if len(item) != 2:
                    print('error item:', repr(line))
                    continue
                multi_s1 = item[0].split('$$$')
                s2 = item[1]
                multi_s1_ids = cls.multi_sent2ids(multi_s1, word2id, max_word_len=50)
                s2_ids = cls.sent2ids(s2, word2id, max_word_len=50)
                if i < 5:  # check
                    print(f'check {i}:')
                    print(f'{multi_s1} -> {multi_s1_ids}')
                    print(f'{s2} -> {s2_ids}')
                data['multi_s1'].append(multi_s1_ids)
                data['s2'].append(s2_ids)
        data['num_data'] = len(data['s2'])
        return data

    @classmethod
    def generate_tfrecord(cls, file, token2id_dct, tfrecord_file):
        from qiznlp.common.tfrecord_utils import items2tfrecord
        word2id = token2id_dct['word2id']

        def items_gen():
            with open(file, 'r', encoding='U8') as f:
                for i, line in enumerate(f):
                    item = line.strip().split('\t')
                    if len(item) != 2:
                        print('error item:', repr(line))
                        continue
                    try:
                        multi_s1 = item[0].split('$$$')
                        s2 = item[1]
                        multi_s1_ids = cls.multi_sent2ids(multi_s1, word2id, max_word_len=50)
                        s2_ids = cls.sent2ids(s2, word2id, max_word_len=50)
                        if i < 5:  # check
                            print(f'check {i}:')
                            print(f'{multi_s1} -> {multi_s1_ids}')
                            print(f'{s2} -> {s2_ids}')
                        d = {
                            'multi_s1': multi_s1_ids,
                            's2': s2_ids,
                        }
                        yield d
                    except Exception as e:
                        print('Exception occur in items_gen()!\n', e)
                        continue

        count = items2tfrecord(items_gen(), tfrecord_file)
        return count

    @classmethod
    def load_tfrecord(cls, tfrecord_file, batch_size=128, index=None, shard=None):
        from qiznlp.common.tfrecord_utils import tfrecord2dataset
        if not os.path.exists(tfrecord_file):
            return None, None
        feat_dct = {
            'multi_s1': tf.VarLenFeature(tf.int64),
            's2': tf.VarLenFeature(tf.int64),
        }
        shape_dct = {
            'multi_s1': [conf.max_turn, -1],
        }
        dataset, count = tfrecord2dataset(tfrecord_file, feat_dct, shape_dct=shape_dct, batch_size=batch_size, auto_pad=True, index=index, shard=shard)
        return dataset, count

    def get_signature_export_model(self):
        inputs_dct = {
            'multi_s1': self.multi_s1,
            'dropout_rate': self.dropout_rate,
        }
        outputs_dct = {
            'decoded_ids': self.decoded_ids,
            'scores': self.scores,
        }
        return inputs_dct, outputs_dct

    @classmethod
    def get_signature_load_pbmodel(cls):
        inputs_lst = ['multi_s1', 'dropout_rate']
        outputs_lst = ['decoded_ids', 'scores']
        return inputs_lst, outputs_lst

    @classmethod
    def from_pbmodel(cls, pbmodel_dir, sess):
        meta_graph_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], pbmodel_dir)  # 从pb模型载入graph和variable,绑定到sess
        signature = meta_graph_def.signature_def  # 取出signature_def
        default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
        inputs_lst, output_lst = cls.get_signature_load_pbmodel()  # 类方法
        pb_dict = {}
        for k in inputs_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].inputs[k].name)  # 从signature中获取输入输出的tensor_name,并从graph中取出
        for k in output_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].outputs[k].name)
        model = cls(build_graph=False)  # 里面不再构造图
        for k, v in pb_dict.items():
            setattr(model, k, v)  # 绑定必要的输入输出到实例
        return model

    @classmethod
    def from_ckpt_meta(cls, ckpt_name, sess, graph):
        with graph.as_default():
            saver = tf.train.import_meta_graph(ckpt_name + '.meta', clear_devices=True)
            sess.run(tf.global_variables_initializer())

        model = cls(build_graph=False)  # 里面不再构造图
        # 绑定必要的输入输出到实例
        model.multi_s1 = graph.get_tensor_by_name('multi_s1:0')
        model.dropout_rate = graph.get_tensor_by_name('dropout_rate:0')
        model.decoded_ids = graph.get_tensor_by_name('decoded_ids:0')
        model.scores = graph.get_tensor_by_name('scores:0')

        saver.restore(sess, ckpt_name)
        print(f':: restore success! {ckpt_name}')
        return model, saver



import tensorflow as tf
from . import operations as op
def similarity(x, y, x_lengths, y_lengths):
    '''calculate similarity with two 3d tensor.

    Args:
        x: a tensor with shape [batch, time_x, dimension]
        y: a tensor with shape [batch, time_y, dimension]

    Returns:
        a tensor with shape [batch, time_x, time_y]

    Raises:
        ValueError: if
            the dimenisons of x and y are not equal.
    '''
    with tf.variable_scope('x_attend_y'):
        try:
            x_a_y = block(
                x, y, y,
                Q_lengths=x_lengths, K_lengths=y_lengths)
        except ValueError:
            tf.get_variable_scope().reuse_variables()
            x_a_y = block(
                x, y, y,
                Q_lengths=x_lengths, K_lengths=y_lengths)

    with tf.variable_scope('y_attend_x'):
        try:
            y_a_x = block(
                y, x, x,
                Q_lengths=y_lengths, K_lengths=x_lengths)
        except ValueError:
            tf.get_variable_scope().reuse_variables()
            y_a_x = block(
                y, x, x,
                Q_lengths=y_lengths, K_lengths=x_lengths)

    return tf.matmul(x + x_a_y, y + y_a_x, transpose_b=True)


def dynamic_L(x):
    '''Attention machanism to combine the infomation, 
       from https://arxiv.org/pdf/1612.01627.pdf.

    Args:
        x: a tensor with shape [batch, time, dimension]

    Returns:
        a tensor with shape [batch, dimension]

    Raises:
    '''
    key_0 = tf.get_variable(
        name='key',
        shape=[x.shape[-1]],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(
            -tf.sqrt(6./tf.cast(x.shape[-1], tf.float32)),
            tf.sqrt(6./tf.cast(x.shape[-1], tf.float32))))

    key = op.dense(x, add_bias=False) #[batch, time, dimension]
    weight = tf.reduce_sum(tf.multiply(key, key_0), axis=-1)  #[batch, time]
    weight = tf.expand_dims(tf.nn.softmax(weight), -1)  #[batch, time, 1]

    L = tf.reduce_sum(tf.multiply(x, weight), axis=1) #[batch, dimension]
    return L 

def loss(x, y, num_classes=2, is_clip=True, clip_value=10):
    '''From info x calculate logits as return loss.

    Args:
        x: a tensor with shape [batch, dimension]
        num_classes: a number

    Returns:
        loss: a tensor with shape [1], which is the average loss of one batch
        logits: a tensor with shape [batch, 1]

    Raises:
        AssertionError: if
            num_classes is not a int greater equal than 2.
    TODO:
        num_classes > 2 may be not adapted.
    '''
    assert isinstance(num_classes, int)
    assert num_classes >= 2

    W = tf.get_variable(
        name='weights',
        shape=[x.shape[-1], num_classes-1],
        initializer=tf.orthogonal_initializer())
    bias = tf.get_variable(
        name='bias',
        shape=[num_classes-1],
        initializer=tf.zeros_initializer())

    logits = tf.reshape(tf.matmul(x, W) + bias, [-1])
    loss = tf.nn.sigmoid_cross_entropy_with_logits(
        labels=tf.cast(y, tf.float32),
        logits=logits)
    loss = tf.reduce_mean(tf.clip_by_value(loss, -clip_value, clip_value))

    return loss, logits

def attention(
    Q, K, V, 
    Q_lengths, K_lengths, 
    attention_type='dot', 
    is_mask=True, mask_value=-2**32+1,
    drop_prob=None):
    '''Add attention layer.
    Args:
        Q: a tensor with shape [batch, Q_time, Q_dimension]
        K: a tensor with shape [batch, time, K_dimension]
        V: a tensor with shape [batch, time, V_dimension]

        Q_length: a tensor with shape [batch]
        K_length: a tensor with shape [batch]

    Returns:
        a tensor with shape [batch, Q_time, V_dimension]

    Raises:
        AssertionError: if
            Q_dimension not equal to K_dimension when attention type is dot.
    '''
    assert attention_type in ('dot', 'bilinear')
    if attention_type == 'dot':
        assert Q.shape[-1] == K.shape[-1]

    Q_time = tf.shape(Q)[1]
    K_time = tf.shape(K)[1]

    if attention_type == 'dot':
        logits = op.dot_sim(Q, K) #[batch, Q_time, time]
    if attention_type == 'bilinear':
        logits = op.bilinear_sim(Q, K)

    if is_mask:
        mask = op.mask(Q_lengths, K_lengths, Q_time, K_time) #[batch, Q_time, K_time]
        logits = mask * logits + (1 - mask) * mask_value
    
    attention = tf.nn.softmax(logits)

    if drop_prob is not None:
        # print('use attention drop')
        attention = tf.layers.dropout(attention, rate=drop_prob)

    return op.weighted_sum(attention, V)

def FFN(x, out_dimension_0=None, out_dimension_1=None):
    '''Add two dense connected layer, max(0, x*W0+b0)*W1+b1.

    Args:
        x: a tensor with shape [batch, time, dimension]
        out_dimension: a number which is the output dimension

    Returns:
        a tensor with shape [batch, time, out_dimension]

    Raises:
    '''
    with tf.variable_scope('FFN_1'):
        y = op.dense(x, out_dimension_0)
        y = tf.nn.relu(y)
    with tf.variable_scope('FFN_2'):
        z = op.dense(y, out_dimension_1) #, add_bias=False)  #!!!!
    return z

def block(
    Q, K, V, 
    Q_lengths, K_lengths, 
    attention_type='dot', 
    is_layer_norm=True, 
    is_mask=True, mask_value=-2**32+1,
    drop_prob=None):
    '''Add a block unit from https://arxiv.org/pdf/1706.03762.pdf.
    Args:
        Q: a tensor with shape [batch, Q_time, Q_dimension]
        K: a tensor with shape [batch, time, K_dimension]
        V: a tensor with shape [batch, time, V_dimension]

        Q_length: a tensor with shape [batch]
        K_length: a tensor with shape [batch]

    Returns:
        a tensor with shape [batch, time, dimension]

    Raises:
    '''
    att = attention(Q, K, V, 
                    Q_lengths, K_lengths, 
                    attention_type=attention_type, 
                    is_mask=is_mask, mask_value=mask_value,
                    drop_prob=drop_prob)
    # attn [batch, q_len, hid]
    # K_lengths [batch]
    tmp = tf.expand_dims(tf.expand_dims(tf.cast(tf.not_equal(K_lengths, 0), tf.float32), -1), -1)  # [batch,1,1]
    att = att * tmp
    
    if is_layer_norm:
        with tf.variable_scope('attention_layer_norm'):
            y = op.layer_norm_debug(Q + att)
    else:
        y = Q + att

    z = FFN(y)
    if is_layer_norm:
        with tf.variable_scope('FFN_layer_norm'):
            w = op.layer_norm_debug(y + z)
    else:
        w = y + z
    return w

def CNN(x, out_channels, filter_size, pooling_size, add_relu=True):
    '''Add a convlution layer with relu and max pooling layer.

    Args:
        x: a tensor with shape [batch, in_height, in_width, in_channels]
        out_channels: a number
        filter_size: a number
        pooling_size: a number

    Returns:
        a flattened tensor with shape [batch, num_features]

    Raises:
    '''
    #calculate the last dimension of return
    num_features = ((tf.shape(x)[1]-filter_size+1)/pooling_size * 
        (tf.shape(x)[2]-filter_size+1)/pooling_size) * out_channels

    in_channels = x.shape[-1]
    weights = tf.get_variable(
        name='filter',
        shape=[filter_size, filter_size, in_channels, out_channels],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    bias = tf.get_variable(
        name='bias',
        shape=[out_channels],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    conv = tf.nn.conv2d(x, weights, strides=[1, 1, 1, 1], padding="VALID")
    conv = conv + bias

    if add_relu:
        conv = tf.nn.relu(conv)

    pooling = tf.nn.max_pool(
        conv, 
        ksize=[1, pooling_size, pooling_size, 1],
        strides=[1, pooling_size, pooling_size, 1], 
        padding="VALID")

    return tf.contrib.layers.flatten(pooling)

def CNN_3d(x, out_channels_0, out_channels_1, add_relu=True):
    '''Add a 3d convlution layer with relu and max pooling layer.

    Args:
        x: a tensor with shape [batch, in_depth, in_height, in_width, in_channels]
        out_channels: a number
        filter_size: a number
        pooling_size: a number

    Returns:
        a flattened tensor with shape [batch, num_features]

    Raises:
    '''
    in_channels = x.shape[-1]
    weights_0 = tf.get_variable(
        name='filter_0',
        shape=[3, 3, 3, in_channels, out_channels_0],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    bias_0 = tf.get_variable(
        name='bias_0',
        shape=[out_channels_0],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    conv_0 = tf.nn.conv3d(x, weights_0, strides=[1, 1, 1, 1, 1], padding="SAME")
    print('conv_0 shape: %s' %conv_0.shape)
    conv_0 = conv_0 + bias_0

    if add_relu:
        conv_0 = tf.nn.elu(conv_0)

    pooling_0 = tf.nn.max_pool3d(
        conv_0, 
        ksize=[1, 3, 3, 3, 1],
        strides=[1, 3, 3, 3, 1], 
        padding="SAME")
    print('pooling_0 shape: %s' %pooling_0.shape)

    #layer_1
    weights_1 = tf.get_variable(
        name='filter_1',
        shape=[3, 3, 3, out_channels_0, out_channels_1],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    bias_1 = tf.get_variable(
        name='bias_1',
        shape=[out_channels_1],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    conv_1 = tf.nn.conv3d(pooling_0, weights_1, strides=[1, 1, 1, 1, 1], padding="SAME")
    print('conv_1 shape: %s' %conv_1.shape)
    conv_1 = conv_1 + bias_1

    if add_relu:
        conv_1 = tf.nn.elu(conv_1)

    pooling_1 = tf.nn.max_pool3d(
        conv_1, 
        ksize=[1, 3, 3, 3, 1],
        strides=[1, 3, 3, 3, 1], 
        padding="SAME")
    print('pooling_1 shape: %s' %pooling_1.shape)

    return tf.contrib.layers.flatten(pooling_1)

def CNN_3d_2d(x, out_channels_0, out_channels_1, add_relu=True):
    '''Add a 3d convlution layer with relu and max pooling layer.

    Args:
        x: a tensor with shape [batch, in_depth, in_height, in_width, in_channels]
        out_channels: a number
        filter_size: a number
        pooling_size: a number

    Returns:
        a flattened tensor with shape [batch, num_features]

    Raises:
    '''
    in_channels = x.shape[-1]
    weights_0 = tf.get_variable(
        name='filter_0',
        shape=[1, 3, 3, in_channels, out_channels_0],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    bias_0 = tf.get_variable(
        name='bias_0',
        shape=[out_channels_0],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    conv_0 = tf.nn.conv3d(x, weights_0, strides=[1, 1, 1, 1, 1], padding="SAME")
    print('conv_0 shape: %s' %conv_0.shape)
    conv_0 = conv_0 + bias_0

    if add_relu:
        conv_0 = tf.nn.elu(conv_0)

    pooling_0 = tf.nn.max_pool3d(
        conv_0, 
        ksize=[1, 1, 3, 3, 1],
        strides=[1, 1, 3, 3, 1], 
        padding="SAME")
    print('pooling_0 shape: %s' %pooling_0.shape)

    #layer_1
    weights_1 = tf.get_variable(
        name='filter_1',
        shape=[1, 3, 3, out_channels_0, out_channels_1],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    bias_1 = tf.get_variable(
        name='bias_1',
        shape=[out_channels_1],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    conv_1 = tf.nn.conv3d(pooling_0, weights_1, strides=[1, 1, 1, 1, 1], padding="SAME")
    print('conv_1 shape: %s' %conv_1.shape)
    conv_1 = conv_1 + bias_1

    if add_relu:
        conv_1 = tf.nn.elu(conv_1)

    pooling_1 = tf.nn.max_pool3d(
        conv_1, 
        ksize=[1, 1, 3, 3, 1],
        strides=[1, 1, 3, 3, 1], 
        padding="SAME")
    print('pooling_1 shape: %s' %pooling_1.shape)

    return tf.contrib.layers.flatten(pooling_1)

def CNN_3d_change(x, out_channels_0, out_channels_1, add_relu=True):
    '''Add a 3d convlution layer with relu and max pooling layer.

    Args:
        x: a tensor with shape [batch, in_depth, in_height, in_width, in_channels]
        out_channels: a number
        filter_size: a number
        pooling_size: a number

    Returns:
        a flattened tensor with shape [batch, num_features]

    Raises:
    '''
    in_channels = x.shape[-1]
    weights_0 = tf.get_variable(
        name='filter_0',
        shape=[3, 3, 3, in_channels, out_channels_0],
        dtype=tf.float32,
        #initializer=tf.random_normal_initializer(0, 0.05))
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    bias_0 = tf.get_variable(
        name='bias_0',
        shape=[out_channels_0],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())
    #Todo
    g_0 = tf.get_variable(name='scale_0',
        shape = [out_channels_0],
        dtype=tf.float32,
        initializer=tf.ones_initializer())
    weights_0 = tf.reshape(g_0, [1, 1, 1, out_channels_0]) * tf.nn.l2_normalize(weights_0, [0, 1, 2])

    conv_0 = tf.nn.conv3d(x, weights_0, strides=[1, 1, 1, 1, 1], padding="VALID")
    print('conv_0 shape: %s' %conv_0.shape)
    conv_0 = conv_0 + bias_0
    #######
    '''
    with tf.variable_scope('layer_0'):
        conv_0 = op.layer_norm(conv_0, axis=[1, 2, 3, 4])
        print('layer_norm in cnn')
    '''
    if add_relu:
        conv_0 = tf.nn.elu(conv_0)

    pooling_0 = tf.nn.max_pool3d(
        conv_0, 
        ksize=[1, 2, 3, 3, 1],
        strides=[1, 2, 3, 3, 1], 
        padding="VALID")
    print('pooling_0 shape: %s' %pooling_0.shape)

    #layer_1
    weights_1 = tf.get_variable(
        name='filter_1',
        shape=[2, 2, 2, out_channels_0, out_channels_1],
        dtype=tf.float32,
        initializer=tf.random_uniform_initializer(-0.01, 0.01))
    
    bias_1 = tf.get_variable(
        name='bias_1',
        shape=[out_channels_1],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())
    
    g_1 = tf.get_variable(name='scale_1',
        shape = [out_channels_1],
        dtype=tf.float32,
        initializer=tf.ones_initializer())
    weights_1 = tf.reshape(g_1, [1, 1, 1, out_channels_1]) * tf.nn.l2_normalize(weights_1, [0, 1, 2])

    conv_1 = tf.nn.conv3d(pooling_0, weights_1, strides=[1, 1, 1, 1, 1], padding="VALID")
    print('conv_1 shape: %s' %conv_1.shape)
    conv_1 = conv_1 + bias_1
    #with tf.variable_scope('layer_1'):
    #    conv_1 = op.layer_norm(conv_1, axis=[1, 2, 3, 4])

    if add_relu:
        conv_1 = tf.nn.elu(conv_1)

    pooling_1 = tf.nn.max_pool3d(
        conv_1, 
        ksize=[1, 3, 3, 3, 1],
        strides=[1, 3, 3, 3, 1], 
        padding="VALID")
    print('pooling_1 shape: %s' %pooling_1.shape)

    return tf.contrib.layers.flatten(pooling_1)

def RNN_last_state(x, lengths, hidden_size):
    '''encode x with a gru cell and return the last state.
    
    Args:
        x: a tensor with shape [batch, time, dimension]
        length: a tensor with shape [batch]

    Return:
        a tensor with shape [batch, hidden_size]

    Raises:
    '''
    cell = tf.nn.rnn_cell.GRUCell(hidden_size)
    outputs, last_states = tf.nn.dynamic_rnn(cell, x, lengths, dtype=tf.float32)
    return outputs, last_states





import os, sys, re
import time
import jieba
import numpy as np
import tensorflow as tf

import qiznlp.common.utils as utils
import qiznlp.common.train_helper as train_helper
from qiznlp.run.run_base import Run_Model_Base, check_and_update_param_of_model_pyfile

curr_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(curr_dir + '/..')  # 添加上级目录即默认qiznlp根目录
from model.multi_mch_model import Model as MMCH_Model

try:
    import horovod.tensorflow as hvd
    # 示例：horovodrun -np 2 -H localhost:2 python run_s2s.py
    # 可根据local_rank设置shard的数据，保证各个gpu采样的数据不重叠。
except:
    HVD_ENABLE = False
else:
    HVD_ENABLE = True

conf = utils.dict2obj({
    'early_stop_patience': None,
    'just_save_best': True,
    'n_epochs': 10,
    'data_type': 'tfrecord',
    # 'data_type': 'pkldata',
})


class Run_Model_MMCH(Run_Model_Base):
    def __init__(self, model_name, tokenize=None, pbmodel_dir=None, use_hvd=False):
        # 维护sess graph config saver
        self.model_name = model_name
        if tokenize is None:
            self.jieba = jieba.Tokenizer()
            # self.jieba.load_userdict(f'{curr_dir}/../data/segword.dct')
            self.tokenize = lambda t: self.jieba.lcut(re.sub(r'\s+', '，', t))
        else:
            self.tokenize = tokenize
        self.cut = lambda t: ' '.join(self.tokenize(t))
        self.token2id_dct = {
            # 'word2id': utils.Any2Id.from_file(f'{curr_dir}/../data/mmch_word2id.dct', use_line_no=True),  # 自有数据
            # 'word2id': utils.Any2Id.from_file(f'{curr_dir}/../data/mmch_char2id.dct', use_line_no=True),  # 自有数据
            'word2id': utils.Any2Id.from_file(f'{curr_dir}/../data/DB_mmch_word2id.dct', use_line_no=True),  # 豆瓣多轮语料
            'char2id': utils.Any2Id.from_file(f'{curr_dir}/../data/DB_mmch_char2id.dct', use_line_no=True),  # 豆瓣多轮语料
        }
        self.config = tf.ConfigProto(allow_soft_placement=True,
                                     gpu_options=tf.GPUOptions(allow_growth=True),
                                     )
        self.use_hvd = use_hvd if HVD_ENABLE else False
        if self.use_hvd:
            hvd.init()
            self.hvd_rank = hvd.rank()
            self.hvd_size = hvd.size()
            self.config.gpu_options.visible_device_list = str(hvd.local_rank())
        self.graph = tf.Graph()
        self.sess = tf.Session(config=self.config, graph=self.graph)

        if pbmodel_dir is not None:  # 只能做predict
            self.model = MMCH_Model.from_pbmodel(pbmodel_dir, self.sess)
        else:
            with self.graph.as_default():
                self.model = MMCH_Model(model_name=self.model_name, run_model=self)
                if self.use_hvd:
                    self.model.optimizer._lr = self.model.optimizer._lr * self.hvd_size  # 分布式训练大batch增大学习率
                    self.model.hvd_optimizer = hvd.DistributedOptimizer(self.model.optimizer)
                    self.model.train_op = self.model.hvd_optimizer.minimize(self.model.loss, global_step=self.model.global_step)
                self.sess.run(tf.global_variables_initializer())
                if self.use_hvd:
                    self.sess.run(hvd.broadcast_global_variables(0))

        with self.graph.as_default():
            self.saver = tf.train.Saver(max_to_keep=100)  # must in the graph context

    def train_step(self, feed_dict):
        _, step, loss, accuracy = self.sess.run([self.model.train_op,
                                                 self.model.global_step,
                                                 self.model.loss,
                                                 self.model.accuracy,
                                                 ],
                                                feed_dict=feed_dict)
        return step, loss, accuracy

    def eval_step(self, feed_dict):
        loss, accuracy, y_prob = self.sess.run([self.model.loss,
                                                self.model.accuracy,
                                                self.model.y_prob,
                                                ],
                                               feed_dict=feed_dict)
        return loss, accuracy, y_prob

    def train(self, ckpt_dir, raw_data_file, preprocess_raw_data, batch_size = 100, save_data_prefix = None):
        save_data_prefix = os.path.basename(ckpt_dir) if save_data_prefix is None else save_data_prefix
        train_epo_steps, dev_epo_steps, test_epo_steps, gen_feed_dict = self.prepare_data(conf.data_type,
                                                                                          raw_data_file,
                                                                                          preprocess_raw_data,
                                                                                          batch_size,
                                                                                          save_data_prefix=save_data_prefix,
                                                                                          update_txt=False,
                                                                                          )
        self.is_master = True
        if hasattr(self, 'hvd_rank') and self.hvd_rank != 0:  # 分布式训练且非master
            dev_epo_steps, test_epo_steps = None, None  # 不进行验证和测试
            self.is_master = False

        # 字典大小自动对齐
        check_and_update_param_of_model_pyfile({
            'vocab_size': (self.model.conf.vocab_size, len(self.token2id_dct['word2id'])),
            'char_vocab_size': (self.model.conf.char_vocab_size, len(self.token2id_dct['char2id'])),
        }, self.model)

        train_info = {}
        for epo in range(1, 1 + conf.n_epochs):
            train_info[epo] = {}

            # train
            time0 = time.time()
            epo_num_example = 0
            trn_epo_loss = []
            trn_epo_acc = []
            for i in range(train_epo_steps):
                feed_dict = gen_feed_dict(i, epo, 'train')
                epo_num_example += feed_dict.pop('num')

                step_start_time = time.time()
                step, loss, acc = self.train_step(feed_dict)
                trn_epo_loss.append(loss)
                trn_epo_acc.append(acc)

                if self.is_master:
                    print(f'\repo:{epo} step:{i + 1}/{train_epo_steps} num:{epo_num_example} '
                          f'cur_loss:{loss:.3f} epo_loss:{np.mean(trn_epo_loss):.3f} '
                          f'epo_acc:{np.mean(trn_epo_acc):.3f} '
                          f'sec/step:{time.time() - step_start_time:.2f}',
                          end=f'{os.linesep if i == train_epo_steps - 1 else ""}',
                          )

            trn_loss = np.mean(trn_epo_loss)
            trn_acc = np.mean(trn_epo_acc)
            if self.is_master:
                print(f'epo:{epo} trn loss {trn_loss:.3f} '
                      f'trn acc {trn_acc:.3f} '
                      f'elapsed {(time.time() - time0) / 60:.2f} min')
            train_info[epo]['trn_loss'] = trn_loss
            train_info[epo]['trn_acc'] = trn_acc

            if not self.is_master:
                continue

            # dev or test
            for mode in ['dev', 'test']:
                epo_steps = {'dev': dev_epo_steps, 'test': test_epo_steps}[mode]
                if epo_steps is None:
                    continue
                time0 = time.time()
                epo_loss = []
                epo_acc = []
                # to calc recall
                epo_s1 = []
                epo_prob = []
                epo_y = []
                for i in range(epo_steps):
                    feed_dict = gen_feed_dict(i, epo, mode)
                    loss, acc, y_prob = self.eval_step(feed_dict)

                    epo_loss.append(loss)
                    epo_acc.append(acc)

                    # to calc recall
                    epo_s1.extend(feed_dict[self.model.multi_s1])
                    epo_prob.extend(y_prob)
                    epo_y.extend(feed_dict[self.model.target])

                loss = np.mean(epo_loss)
                acc = np.mean(epo_acc)
                recall = train_helper.calc_recall(epo_s1, epo_prob, epo_y, strip_pad=True)  # calc recall
                recall = list(map(lambda e: round(e, 3), recall))

                print(f'epo:{epo} {mode} loss {loss:.3f} '
                      f'{mode} acc {acc:.3f} '
                      f'{mode} recall@n {recall} '  # print recall
                      f'elapsed {(time.time() - time0) / 60:.2f} min')
                train_info[epo][f'{mode}_loss'] = loss
                train_info[epo][f'{mode}_acc'] = acc
                train_info[epo][f'{mode}_recall@n'] = recall

            info_str = f'{trn_loss:.2f}-{train_info[epo]["dev_loss"]:.2f}'
            info_str += f'-{trn_acc:.3f}-{train_info[epo]["dev_acc"]:.3f}'
            info_str += f'-{train_info[epo]["dev_recall@n"][0]:.3f}'

            if conf.just_save_best:
                if self.should_save(epo, train_info, 'dev_recall@n', greater_is_better=True):
                    self.delete_ckpt(ckpt_dir=ckpt_dir)  # 删掉已存在的
                    self.save(ckpt_dir=ckpt_dir, epo=epo, info_str=info_str)
            else:
                self.save(ckpt_dir=ckpt_dir, epo=epo, info_str=info_str)

            utils.obj2json(train_info, f'{ckpt_dir}/metrics.json')
            print('=' * 40, end='\n')
            if conf.early_stop_patience:
                if self.stop_training(conf.early_stop_patience, train_info, 'dev_loss', greater_is_better=False):
                    print('early stop training!')
                    print('train_info', train_info)
                    break

    def predict(self, multi_s1_lst, s2_lst, need_cut=True, batch_size=100):
        assert len(multi_s1_lst) == len(s2_lst)
        if need_cut:
            multi_s1_lst = ['$$$'.join([self.cut(s1) for s1 in multi_s1.split('$$$')]) for multi_s1 in multi_s1_lst]
            s2_lst = [self.cut(s2) for s2 in s2_lst]

        pred_lst = []
        for i in range(0, len(multi_s1_lst), batch_size):
            batch_multi_s1 = multi_s1_lst[i:i + batch_size]
            batch_s2 = s2_lst[i:i + batch_size]
            feed_dict = self.model.create_feed_dict_from_raw(batch_multi_s1, batch_s2, [], self.token2id_dct, mode='infer')
            probs = self.sess.run(self.model.y_prob, feed_dict)  # [batch]
            threshold = 2.5 if self.model_name.startswith('MRFN') else 0.5
            preds = [1 if prob >= threshold else 0 for prob in probs]
            pred_lst.extend(preds)
        return pred_lst


def preprocess_raw_data(file, tokenize, token2id_dct, **kwargs):
    """
    # 处理自有数据函数模板
    # file文件数据格式: 多轮对话句子1\t多轮对话句子2\t...\t多轮对话句子n
    # [filter] 过滤
    # [segment] 分词
    # [build vocab] 构造词典
    # [split] train-dev-test
    """
    seg_file = file.rsplit('.', 1)[0] + '_seg.txt'
    if not os.path.exists(seg_file):
        sess_lst = utils.file2items(file)
        # 过滤
        # filter here

        print('过滤后数据量', len(sess_lst))

        # 分词
        for i, sess in enumerate(sess_lst):
            # sess_lst[i] = [' '.join(s) for s in sess]  # 按字分
            sess_lst[i] = [' '.join(tokenize(s)) for s in sess]  # 按词分
        utils.list2file(seg_file, sess_lst)
        print('保存分词后数据成功', '数据量', len(sess_lst), seg_file)
    else:
        # 读取分词好的数据
        sess_lst = utils.file2items(seg_file)

    # 转为多轮格式 multi-turn之间用$$$分隔
    items = []
    for sess in sess_lst:
        for i in range(1, len(sess)):
            multi_src = '$$$'.join(sess[:i])
            tgt = sess[i]
            items.append([multi_src, tgt])
    # items: [['w w w$$$w w', 'w w w'],...]

    # 划分 不分测试集
    train_items, dev_items = utils.split_file(items, ratio='19:1', shuffle=True, seed=1234)

    # 构造词典(option) 字词联合
    need_to_rebuild = []
    for token2id_name in token2id_dct:
        if not token2id_dct[token2id_name]:
            print(f'字典{token2id_name} 载入不成功, 将生成并保存')
            need_to_rebuild.append(token2id_name)

    if need_to_rebuild:
        print(f'生成缺失词表文件...{need_to_rebuild}')
        for items in [train_items, dev_items]:  # 字典只统计train和dev
            for item in items:
                if 'word2id' in need_to_rebuild:
                    for sent in item[0].split('$$$'):
                        token2id_dct['word2id'].to_count(sent.split(' '))
                    token2id_dct['word2id'].to_count(item[1].split(' '))
                if 'char2id' in need_to_rebuild:
                    for sent in item[0].split('$$$'):
                        token2id_dct['char2id'].to_count(list(sent.replace(' ', '')))
                    token2id_dct['char2id'].to_count(list(item[1].replace(' ', '')))
        if 'word2id' in need_to_rebuild:
            token2id_dct['word2id'].rebuild_by_counter(restrict=['<pad>', '<unk>', '<eos>'], min_freq=1, max_vocab_size=30000)
            token2id_dct['word2id'].save(f'{curr_dir}/../data/mmch_word2id.dct')
        if 'char2id' in need_to_rebuild:
            token2id_dct['char2id'].rebuild_by_counter(restrict=['<pad>', '<unk>', '<eos>'], min_freq=1, max_vocab_size=4000)
            token2id_dct['char2id'].save(f'{curr_dir}/../data/mmch_char2id.dct')
    else:
        print('使用已有词表文件...')

    # 负采样
    train_items = train_helper.gen_pos_neg_sample(train_items, sample_idx=1, num_neg_exm=4)
    dev_items = train_helper.gen_pos_neg_sample(dev_items, sample_idx=1, num_neg_exm=4)
    return train_items, dev_items, None


def preprocess_common_dataset_Douban(file, tokenize, token2id_dct, **kwargs):
    # 豆瓣多轮语料
    # 对话数据分词,且生成词和字的字典word2id char2id
    file = f'{curr_dir}/../data/Douban_Sess662.txt'

    def Doubanchange2items(file):
        # 转为[multi_src, tgt]格式
        # 分词
        seg_file = file.rsplit('.', 1)[0] + '_seg.txt'
        if not os.path.exists(seg_file):
            items = utils.file2items(file)
            # 分词
            for i, item in enumerate(items):
                for j in range(len(item)):
                    items[i][j] = ' '.join(tokenize(items[i][j]))
            utils.list2file(seg_file, items)
            print('保存分词后数据成功', '数据量', len(items), seg_file)
        else:
            items = utils.file2items(seg_file)

        exm_lst = []
        sess_lst = items
        for sess in sess_lst:
            for i in range(1, len(sess)):
                multi_src = '$$$'.join(sess[:i])
                tgt = sess[i]
                exm_lst.append([multi_src, tgt])
        return exm_lst

    items = Doubanchange2items(file)  # [['w w w$$$w w', 'w w w'],...]

    # 划分 不分测试集
    train_items, dev_items = utils.split_file(items, ratio='19:1', shuffle=True, seed=1234)

    # 构造词典(option) 字词联合
    need_to_rebuild = []
    for token2id_name in token2id_dct:
        if not token2id_dct[token2id_name]:
            print(f'字典{token2id_name} 载入不成功, 将生成并保存')
            need_to_rebuild.append(token2id_name)

    if need_to_rebuild:
        print(f'生成缺失词表文件...{need_to_rebuild}')
        for items in [train_items, dev_items]:  # 字典只统计train和dev
            for item in items:
                if 'word2id' in need_to_rebuild:
                    for sent in item[0].split('$$$'):
                        token2id_dct['word2id'].to_count(sent.split(' '))
                    token2id_dct['word2id'].to_count(item[1].split(' '))
                if 'char2id' in need_to_rebuild:
                    for sent in item[0].split('$$$'):
                        token2id_dct['char2id'].to_count(list(sent.replace(' ', '')))
                    token2id_dct['char2id'].to_count(list(item[1].replace(' ', '')))
        if 'word2id' in need_to_rebuild:
            token2id_dct['word2id'].rebuild_by_counter(restrict=['<pad>', '<unk>', '<eos>'], min_freq=1, max_vocab_size=30000)
            token2id_dct['word2id'].save(f'{curr_dir}/../data/DB_mmch_word2id.dct')
        if 'char2id' in need_to_rebuild:
            token2id_dct['char2id'].rebuild_by_counter(restrict=['<pad>', '<unk>', '<eos>'], min_freq=1, max_vocab_size=4000)
            token2id_dct['char2id'].save(f'{curr_dir}/../data/DB_mmch_char2id.dct')
    else:
        print('使用已有词表文件...')

    # 负采样
    train_items = train_helper.gen_pos_neg_sample(train_items, sample_idx=1, num_neg_exm=4)
    dev_items = train_helper.gen_pos_neg_sample(dev_items, sample_idx=1, num_neg_exm=4)
    return train_items, dev_items, None


if __name__ == '__main__':
    os.environ['CUDA_VISIBLE_DEVICES'] = '0'  # 使用CPU设为'-1'

    # rm_mmch = Run_Model_MMCH('DAM')  # use DAM
    rm_mmch = Run_Model_MMCH('MRFN')  # use MRFN

    # 训练自有数据
    # rm_mmch.train('multi_mch_ckpt_1', '../data/multi_mch_example_data.txt', preprocess_raw_data=preprocess_raw_data, batch_size=512)  # train

    # 训练豆瓣多轮语料
    rm_mmch.train('multi_mch_ckpt_DB1', '', preprocess_raw_data=preprocess_common_dataset_Douban, batch_size=128)  # train

    # demo豆瓣多轮检索式对话模型
    rm_mmch.restore('multi_mch_ckpt_DB1')  # for infer
    import readline
    while True:
        try:
            inp = input('enter:($$$分隔多轮句子,|||分隔问句与答句)')
            sent1, sent2 = inp.split('|||')
            need_cut = False if ' ' in sent1 else True
            time0 = time.time()
            ret = rm_mmch.predict([sent1], [sent2], need_cut=need_cut)
            print(ret[0])
            print('elapsed:', time.time() - time0)
        except KeyboardInterrupt:
            exit(0)
        except Exception as e:
            print(e)



#!/usr/bin/env python
# coding: utf-8
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='QizNLP',
    version='0.1.3',
    author='Qznan',
    author_email='summerzynqz@gmail.com',
    description='Quick run NLP in many task',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MPLv2.0',
    url='https://github.com/Qznan/QizNLP',
    packages=find_packages(),
    package_data={'qiznlp':
                      ['data/*.txt',
                       'common/modules/bert/chinese_L-12_H-768_A-12/*.json',
                       'common/modules/bert/chinese_L-12_H-768_A-12/*.txt',
                       ],
                  },
    install_requires=[
        'jieba',
        'tensorflow>=1.8, <=1.14'
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'qiznlp_init=qiznlp:qiznlp_init',
        ]
    },
    keywords='NLP Classification Match Sequence_Label Senquence_to_Senquence Neural_Network',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



#!/usr/bin/env python
# coding=utf-8
import tensorflow as tf


class IDCNN():
    def __init__(self, **kwargs):
        """
        """
        self.kernel_size = 3  # kernel_size
        self.num_filters = 100  # out_channel
        self.repeat_times = 4  # 共4个block
        self.layers = [  # 每个block结构
            {
                'dilation': 1
            },
            {
                'dilation': 3
            },
            {
                'dilation': 5
            },
        ]

    def __call__(self, embedding, name='idcnn', reuse=tf.AUTO_REUSE, **kwargs):
        """
        :param idcnn_inputs embedding: [batch, len, embed] 
        :return: [batch, len, num_filter * repeat_times]
        """
        with tf.variable_scope(name, reuse=reuse):
            inputs = tf.expand_dims(embedding, 1)  # [batch, 1, length, embed] 

            # shape of input = [batch, in_height, in_width, in_channels]
            layerinput = tf.layers.conv2d(inputs, self.num_filters, [1, self.kernel_size],
                                          strides=[1, 1], padding='SAME',
                                          use_bias=False,
                                          kernel_initializer=tf.contrib.layers.xavier_initializer(),
                                          name='init_layer')

            final_out_from_layers = []
            total_width_for_last_dim = 0
            for j in range(self.repeat_times):  # 多个block共享参数
                for i in range(len(self.layers)):
                    dilation = self.layers[i]['dilation']
                    with tf.variable_scope("atrous-conv-layer-%d" % i, reuse=tf.AUTO_REUSE):
                        w = tf.get_variable("filterW", shape=[1, self.kernel_size, self.num_filters, self.num_filters],
                                            initializer=tf.contrib.layers.xavier_initializer())
                        b = tf.get_variable("filterB", shape=[self.num_filters])
                        layeroutput = tf.nn.atrous_conv2d(layerinput,
                                                          w,
                                                          rate=dilation,
                                                          padding="SAME")

                    layeroutput = tf.nn.bias_add(layeroutput, b)
                    layeroutput = tf.nn.relu(layeroutput)
                    if i == (len(self.layers) - 1):
                        final_out_from_layers.append(layeroutput)
                        total_width_for_last_dim += self.num_filters
                    layerinput = layeroutput
            idcnn_output = tf.concat(axis=3, values=final_out_from_layers)  # 【batch,1,len,hid]】
            idcnn_output = tf.squeeze(idcnn_output, 1)

            return idcnn_output  # [batch,len,hid]
